/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1996 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	cfi.c (Common Filter Interface)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	960130	created
	960203	renamed from approxy (application protocol translation library)
//////////////////////////////////////////////////////////////////////#*/
#include <stdio.h>
#include "ystring.h"
#include <ctype.h>

extern int simple_relayf();
extern char *RFC822_readHeader();
extern char *findField();
extern char *nextField();
extern char *matchFields();
extern char *getFieldValue2();
#define getFieldValue(str,fld,buf,siz) getFieldValue2(str,fld,buf,siz)
#define getFV(str,fld,buf)             getFieldValue2(str,fld,buf,sizeof(buf))
#define OPgetFV(spc,pre,fld,buf)  OPgetFieldValue(spc,pre,fld,buf,sizeof(buf))
extern int copyfile1();
extern FILE *TMPFILE();
extern FILE *Gunzip();
extern char *malloc();
extern char *realloc();
#define FTOCL_FIELD	"X-Request"

/*
 * ARGUMENTS
 *
 *	in	 input
 *	out	 output
 *	conninfo information about the client/server connection
 *	convspec specification of conversion
 *
 * RETURN value
 *
 *	-1 error
 *	 0 no translation
 *	 1 translated
 */
cfi(isresp,in,out,conninfo,convspec)
	FILE *in,*out;
	char *conninfo,*convspec;
{	char statline[1024];
	char *head;
	FILE *tmp;
	char clproto[64];
	int withhead;
	int withbody;
	int rcode;

	if( getFV(conninfo,"Client-Protocol",clproto) == 0 )
		strcpy(clproto,"http");

sv1log("## CFI/%s ##\n",clproto);

	if( strcaseeq(clproto,"nntp")
	 || strcaseeq(clproto,"smtp")
	 || strcaseeq(clproto,"pop") )
	{
		filter_msgs(isresp,clproto,in,out,conninfo,convspec);
		return 0;
	}

	/*
	if(!strcaseeq(clproto,"http") ){
	*/
	if( !strncaseeq(clproto,"http",4) ){
		if( non_http_conv(in,out,conninfo,convspec,clproto) ){
			sv1log("## NON-MIME-RELAY\n");
			return 1;
		}else{
			sv1log("## SIMPLE-RELAY\n");
			simple_relayf(in,out);
			return 0;
		}
	}

	/*
	 *	HTTP
	 */
	withhead = 1;
	withbody = 1;
	if( isresp ){
		/*
		 *	Pass through non-HTTP stream
		 */
		if( fPollIn(in,0) < 0 )
			exit(-1);
		if( fread(statline,1,5,in) <= 0 )
			return -1;

		if( strncmp(statline,"HTTP/",5) != 0 ){
			write(fileno(out),statline,5);
			simple_relayf(in,out);
			return 0;
		}
		/*
		 *	Read status line and response header
		 */
		if( fgets(statline+5,sizeof(statline),in) == NULL )
			return -1;

		if( sscanf(statline,"HTTP/%*s %d",&rcode) )
			if( rcode == 304 || rcode == 204 || rcode/100 == 1 ) 
				withbody = 0;
	}else{
		fgets(statline,sizeof(statline),in);
		if( strncmp(statline,"POST",4) != 0
		 && strncmp(statline,"PUT", 3) != 0 )
			withbody = 0;
		if( strstr(statline,"HTTP/") == NULL ){
			withhead = 0;
			withbody = 0;
		}
	}
	if( withhead ){
		head = RFC822_readHeader(in,0);
		if( isresp ){
			char req[4096];
			if( getFV(head,"X-Request",req) )
				if( strncmp(req,"HEAD",4) == 0 )
					withbody = 0;
		}
	}else	head = stralloc("");
	head = realloc(head,strlen(head)+2048); /* make space for rewriting */

	/*
	 *	pass through if not to be filterd
	 */
	if( http_conv(in,out,conninfo,convspec,clproto,statline,head,withbody) == 0 ){
		fputs(statline,out);
		removeFields(head,FTOCL_FIELD,1);
		fputs(head,out);
		simple_relayf(in,out);
		return 0;
	}
	return 1;
}

char *searchSpec(conninfo,specs,statline,head)
	char *conninfo,*specs,*statline,*head;
{	char ictype[1024],iagent[1024],iserver[1024],iencode[1024];
	char request[2048],imethod[256],iurl[2048],iver[256];
	char from[256];
	char orequest[2048],iomethod[32],iourl[2048];
	char rver[256],rstat[256];
	char rcode[32];
	char *dp;
	char clhost[256],svhost[256];
	char clproto[256],svproto[256];
	int top;
	char *spec1,*np;
	char *delp,delc;

	orequest[0] = iourl[0] = 0;
	getFV(head,"X-Request-Original",orequest);
	dp = wordScan(orequest,iomethod);
	dp = wordScan(dp,iourl);

	request[0] = 0;
	getFV(head,"X-Request",request);
	dp = wordScan(request,imethod);
	dp = wordScan(dp,iurl);
	dp = wordScan(dp,iver);

	getFV(head,"X-Request-User-Agent",iagent);
	if( iagent[0] == 0 )
		getFV(head,"User-Agent",iagent);

	dp = wordScan(statline,rver);
	dp = wordScan(dp,rcode);

	getFV(head,"Server",iserver);
	getFV(head,"Content-Type",ictype);
	getFV(head,"Content-Encoding",iencode);
	getFV(conninfo,"Client-Host",clhost);
	getFV(conninfo,"Client-Protocol",clproto);
	getFV(conninfo,"Server-Host",svhost);
	getFV(conninfo,"Server-Protocol",svproto);

	getFV(head,"From",from);
	
	spec1 = specs;
	delp = 0;
	for( spec1 = specs; *spec1; spec1 = np ){
		top = 1;
		if( delp ){ /* restore the delimiter char. */ *delp = delc; }
		for( np = spec1; *np; np++ ){
			if( top ){
				if( *np == '.' && (np[1]=='\r' || np[1] == '\n')
				 || *np == '-' && np[1]=='-' && (np[2]=='\r'||np[2]=='\n')
				){
					if( *np == '-' )
						np += 2;
					else	np += 1;

					while( np[1] == '\r' || np[1] == '\n' )
						np++;
					delp = np;
					delc = *delp;
					*np++ = 0;
					break;
				}
			}
			top = (*np == '\n');
		}

		while( isspace(*spec1) )
			spec1++;
		if( *spec1 == 0 )
			continue;

		sv1vlog("////////////////////////////\n%s\n",spec1);

		if( !matchFields(spec1,"User-Agent",iagent) ) continue;
		if( !matchFields(spec1,"Server",iserver) ) continue;
		if( !matchFields(spec1,"Content-Type",ictype) ) continue;
		if( !matchFields(spec1,"Content-Encoding",iencode) ) continue;
		if( !matchFields(spec1,"Req-Method",imethod) ) continue;
		if( !matchFields(spec1,"Req-Url",iurl) ) continue;
		if( !matchFields(spec1,"Req-Version",iver) ) continue;
		if( !matchFields(spec1,"Req-Url-Orig",iourl) ) continue;

		if( !matchFields(spec1,"X-Request",request) ) continue;
		if( !matchFields(spec1,"X-Request-Method",imethod) ) continue;
		if( !matchFields(spec1,"X-Request-URL",iurl) ) continue;
		if( !matchFields(spec1,"X-Request-Ver",iver) ) continue;
		if( !matchFields(spec1,"X-Request-URL-Orig",iourl) ) continue;

		if( !matchFields(spec1,"Res-Version",rver) ) continue;
		if( !matchFields(spec1,"Res-Status",rstat) ) continue;

		if( !matchFields(spec1,"X-Status",rstat) ) continue;
		if( !matchFields(spec1,"X-Status-Ver",rver) ) continue;
		if( !matchFields(spec1,"X-Status-Code",rcode) ) continue;

		if( !matchFields(spec1,"Client-Host",clhost) ) continue;
		if( !matchFields(spec1,"Client-Protocol",clproto) ) continue;
		if( !matchFields(spec1,"Server-Host",svhost) ) continue;
		if( !matchFields(spec1,"Server-Protocol",svproto) ) continue;
		if( !matchFields(spec1,"From",from) ) continue;

		sv1vlog("///////// MATCHED //////////\n");
		if( delp ){
			spec1 = stralloc(spec1);
		}
		if( delp ){ /* restore the delimiter char. */ *delp = delc; }
		return spec1;
	}
	return NULL;
}

static char *OPgetFieldValue(spec,prefix,fname,value,size)
	char *spec,*prefix,*fname,*value;
{	char *vp,field[256];

	sprintf(field,"%s/%s",prefix,fname);
	if( vp = getFieldValue(spec,field,value,size) )
		return vp;

	sprintf(field,"%s-%s",prefix,fname); /* obsolete ... */
	if( vp = getFieldValue(spec,field,value,size) )
		return vp;

	return NULL;
}

static substitute(fp,conninfo,spec,ftype,fname,statline,head)
	FILE *fp;
	char *conninfo,*spec,*ftype,*fname,*statline,*head;
{	char fvalue[1024];
	char *sp,sc;
	char *hp,hname[1024],hvalue[1024],tmp[1024];

	OPgetFV(spec,ftype,fname,fvalue);
	for( sp = fvalue; sc = *sp; sp++ ){
		if( sc == '#' ){
			while( sp[1] && sp[1] != '\n' )
				sp++;
			continue;
		}else
		if( sc == '$' && sp[1] == '{' ){
			sscanf(sp,"${%[^}]",hname);
			if( sp[2+strlen(hname)] == '}' ){
				if( strcasecmp(hname,"head") == 0 )
					fputs(head,fp);
				else
				if( strncasecmp(hname,"icon.",5) == 0 ){
					getFV(conninfo,"Client-IF-Host",tmp);
					fprintf(fp,"//%s/-/builtin/icons/ysato/%s",tmp,hname+5);
				}else
				if( strcasecmp(hname,"Req-Url-Orig") == 0 ){
					char *url;
					char *dp;

					getFV(head,"X-Request-Original",hvalue);
					url = tmp;
					dp = wordScan(hvalue,tmp);
					dp = wordScan(dp,tmp);
					if( strncmp(url,"/-_-",4) == 0 )
						url += 4;
					fputs(url,fp);
				}else
				if( getFV(conninfo,hname,hvalue) ){
					fputs(hvalue,fp);
				}else
				if( getFV(head,hname,hvalue) )
					fputs(hvalue,fp);
				else{
					fputc(sc,fp);
					continue;
				}
				sp += 2 + strlen(hname);
				continue;
			}
		}
		fputc(sc,fp);
	}
	if( *fvalue )
		fputs("\r\n\r\n",fp);
}

/*
 * read a messge to its logical end
 */
static HTTP_getBody(statline,head,in,out)
	char *statline,*head;
	FILE *in,*out;
{	char cleng[128];
	int leng,rcc,ch,timeout;

	if( strncmp(statline,"POST ",5) != 0 )
		return 0;

	leng = 0;
	if( getFV(head,"Content-Length",cleng) )
		leng = atoi(cleng);

	ch = 0;
	rcc = 0;
	timeout = 3*1000;
	while( 0 < fPollIn(in,timeout) ){
		ch = getc(in);
		if( ch == EOF )
			break;
		putc(ch,out);
		rcc++;
		if( leng <= rcc )
			timeout = 100;
	}

	sv1log("## CFI-getBody[%d/%d] EOF=0x%X\n",rcc,leng,ch&0xFF);
	fflush(out);
	fseek(out,0,0);
	return 1;
}
removeVaryCE(head)
	char *head;
{	char vary[64];

	if( getFieldValue2(head,"Vary",vary,sizeof(vary)) ){
		if( strcaseeq(vary,"Accept-Encoding") )
			removeFields(head,"Vary",0);
	}
}
static http_conv(in,out,conninfo,convspec,clproto,statline,head,withbody)
	FILE *in,*out;
	char *clproto;
	char *conninfo,*convspec,*statline,*head;
{	char *spec;
	char octype[1024],ochset[256],oetype[256],ostatus[256];
	char oheader[1024],prefix[1024],postfix[1024];
	char filter[1024],cgi[1024];
	FILE *intmp,*tmp;
	char *xhead;
	int bodyoff;
	char *ohead,*nhead;
	char *savhead;
	char enc[64];

	spec = searchSpec(conninfo,convspec,statline,head);
	if( spec == NULL )
		return 0;

	savhead = stralloc(head);

	/*
	 *	Rewrite Content-Type field
	 */
	if( OPgetFV(spec,"Output","Content-Type",octype) )
		replaceFieldValue(head,"Content-Type",octype);

	if( OPgetFV(spec,"Output","Charset",ochset) )
		replace_charset(head,ochset);

	if( OPgetFV(spec,"Output","Content-Encoding",oetype) )
		replaceFieldValue(head,"Content-Encoding",oetype);

	if( strncmp(statline,"HTTP/",5) == 0 ){
		int code;
		char ver[32],xstat[1024];
		strcpy(xstat,"X-Status: ");
		linescanX(statline,xstat+strlen(xstat),sizeof(xstat));
		strcat(xstat,"\r\n");
		RFC822_addHeaderField(head,xstat);

		wordScan(statline,ver);
		sscanf(statline,"%*s %d",&code);
		sprintf(xstat,"X-Status-Ver: %s\r\n",ver);
		RFC822_addHeaderField(head,xstat);
		sprintf(xstat,"X-Status-Code: %d\r\n",code);
		RFC822_addHeaderField(head,xstat);
	}

	filterFields(spec,head);

	if( getFieldValue2(head,"Content-Encoding",enc,sizeof(enc)) ){
		FILE *sin = in;
		in = Gunzip(enc,in);
		if( in != sin ){
			removeFields(head,"Content-Encoding",0);
			removeVaryCE(head);
		}
	}

	tmp = TMPFILE("CFI-CONV");

	ohead = nhead = head;
	if( getFV(spec,"Header/Filter",filter)
	 || getFV(spec,"Header-Filter",filter)
	){
		FILE *htmp;
		int nhsize;
		char line[1024],*crlf,crlfb[4];

		htmp = TMPFILE("CFI-Header/Filter");

		if( strncmp(statline,"HTTP/",5) == 0 )
			fprintf(htmp,"Status-Line: %s",statline);
		else
		if( statline[0] && strstr(statline,"HTTP/") ){
			fprintf(htmp,"Request-Line: %s",statline);
			syslog_DEBUG("Request-Line: %s",statline);
		}

		fputs(head,htmp); fflush(htmp); fseek(htmp,0,0);
		System(filter,htmp,tmp);
		fclose(htmp);
		fflush(tmp); fseek(tmp,0,0);
		nhsize = file_size(fileno(tmp));
		nhead = malloc(nhsize+2);
		fread(nhead,1,nhsize,tmp);
		nhead[nhsize] = 0;
		fseek(tmp,0,0);
		Ftruncate(tmp,0,0);
		head = nhead;

		if( getFV(head,"Status-Line",line)
		 || getFV(head,"Request-Line",line) ){
			removeFields(head,"Request-Line",0);
			removeFields(head,"Status-Line",0);
			if( strstr(line,"HTTP/") ){
				if( crlf = strpbrk(statline,"\r\n") )
					strncpy(crlfb,crlf,sizeof(crlfb));
				strcpy(statline,line);
				if( crlf )
					strcat(statline,crlfb);
				syslog_ERROR(">>start-line: %s",statline);
			}
		}
	}

	intmp = NULL; 
	xhead = NULL;

	if( getFV(spec,"CGI",cgi) ){
		char oreq[4096],req[4096];

		intmp = TMPFILE("CFI-CGI");
		getFV(head,"X-Request-Original",oreq);
		getFV(head,"X-Request",req);

		system_CGI(conninfo,oreq,req,head,cgi,in,intmp);
		fseek(intmp,0,0);
		fgets(statline,sizeof(statline),intmp);
		head = xhead = RFC822_readHeader(intmp,0);
		in = intmp;
	}
	if( getFV(spec,"SSI",cgi) ){
		intmp = TMPFILE("CFI-SSI");
/*
		system_SSI(conninfo,in,intmp);
*/
		in = intmp;
	}

	if( OPgetFV(spec,"Output","Status",ostatus) )
		fprintf(tmp,"%s %s\r\n","HTTP/1.0",ostatus);
	else	fputs(statline,tmp);

	if( getFV(spec,"Output-Header",oheader) ){
		char buf[1024];
		genheadf(oheader,buf,sizeof(buf));
		fputs(buf,tmp);
		/*
		fputs(oheader,tmp);
		*/
		fputs("\r\n",tmp);
	}

	removeFields(head,FTOCL_FIELD,1);
	removeFields(head,"X-Status",1);
	fputs(head,tmp);
	fflush(tmp);
	bodyoff = ftell(tmp);

	/*
	 *	Translate
	 */
	if( withbody ){
		if( getFV(spec,"Output-Prefix",prefix) )
			substitute(tmp,conninfo,spec,"Output","Prefix",statline,savhead);

		if( getFV(spec,"Filter",filter)
		 || getFV(spec,"Body-Filter",filter) )
		{	FILE *inbody;

			inbody = TMPFILE("CFI-MESSAGE-BODY");
			if( HTTP_getBody(statline,head,in,inbody) ){
				fclose(in);
				in = inbody;
			}
			System(filter,in,tmp);
		}
		else{
			if( HTTP_getBody(statline,head,in,tmp) ){
			}else{
				copyfile1(in,tmp);
			}
		}

		if( getFV(spec,"Output-Postfix",postfix) )
			substitute(tmp,conninfo,spec,"Output","Postfix",statline,savhead);
	}
	if( getFV(spec,"Message-Filter",filter) ){
		FILE *inmsg;
		inmsg = TMPFILE("CFI-MESSAGE");
		fseek(tmp,0,0);
		System(filter,tmp,inmsg);
		fclose(tmp);
		tmp = inmsg;
	}

/*
	fflush(tmp);
	fseek(tmp,bodyoff,0);
	if( fgets(statline,sizeof(statline),tmp) != NULL ){
		if( strncmp(statline,"CFI/",4) == 0 ){
			char *head;

			sv1log("CFI-CONTROL: %s\n",statline);
			if( head = RFC822_readHeader(tmp,0) ){
				sv1log("CFI-CONTROL: %s\n",head);
				free(head);
			}
		}
	}
*/

	fseek(tmp,0,0);
	if( *savhead && withbody ) /* both head & body exist */
		HTTP_putMIMEmsg(NULL,tmp,out);
	else	copyfile1(tmp,out);
	fflush(out);
	fclose(tmp);
	if( intmp ){
		fclose(intmp);
		free(xhead);
	}
	free(savhead);
	return 1;
}

non_http_conv(in,out,conninfo,convspec,clproto)
	FILE *in,*out;
	char *conninfo,*convspec,*clproto;
{	char *spec,filter[1024];

	spec = searchSpec(conninfo,convspec,"",conninfo);
	if( spec == NULL )
		return 0;

	if( getFV(spec,"Filter",filter)
	 || getFV(spec,"Message-Filter",filter) )
		System(filter,in,out);
	else	simple_relayf(in,out);
	return 1;
}

System(command,in,out)
	char *command;
	FILE *in,*out;
{
	if( 0 <= builtin_filter(NULL,command,in,out,NULL,NULL) )
		return;

	systemFilter(command,in,out);
}

extern char *strSeekEOH();
putMESSAGEline(fp,type,comment)
	FILE *fp;
	char *type,*comment;
{
	return fprintf(fp,"--MESSAGE/%s %s\r\n",type,comment);
}
getMESSAGEline(line,type,comment)
	char *line,*type,*comment;
{
	if( strncmp(line,"--MESSAGE/",10) == 0 )
		return sscanf(line,"--MESSAGE/%s %[^\r\n]",type,comment);
	else	return 0;
}

/*    FILTER_MSGS
 *	Filter a series of mesages which start with a line headed by
 *	  "--MESSAGE/line"
 *	  "--MESSAGE/head"
 *	  "--MESSAGE/body"
 *	  "--MESSAGE/mime"
 *	and end with a line
 *	  "." CRLF
 */

static char *NOHEAD = "";

filter_msgs(isresp,clproto,in,out,conninfo,convspec)
	char *clproto;
	FILE *in,*out;
	char *conninfo,*convspec;
{	FILE *tmp;
	char msgstat[1024];
	int nostat;
	char statline[1024];
	char type[128],comment[128];
	char *head;
	int body;
	char *eoh,eohb[1024],endline[1024];

	tmp = TMPFILE("CFI-MESSAGES");

sv1log("## CFI/MSGS ##\n");

	for(;;){
		if( fPollIn(in,10) == 0 )
			fflush(out);
		if( fgets(msgstat,sizeof(msgstat),in) == NULL )
			break;

sv1log("CFI/MSGS ---- %s",msgstat);

		/*
		 *	out of scoope of the filter
		 */
		if( getMESSAGEline(msgstat,type,comment) == 0 ){
			fputs(msgstat,out);
			continue;
		}

/*
{
	char filter[1024];
	if( getFV(convspec,"Control-Filter",filter) ){
	}
}
*/

		/*
		 *	only status line
		 */
		if( strcmp(type,"line") == 0 ){
			fgets(statline,sizeof(statline),in);
			fputs(statline,out);
			continue;
		}

		/*
		 *	status line and [head][body].CRLF
		 */
		nostat = 0;
		/* if( !isresp ) */
		if( strcaseeq(clproto,"nntp") ){
			if( strncasecmp(comment,"POST",4) == 0 )
				nostat = 1;
		}else
		if( strcaseeq(clproto,"smtp") ){
			if( strncasecmp(comment,"DATA",4) == 0 )
				nostat = 1;
		}
		if( nostat )
			statline[0] = 0;
		else	fgets(statline,sizeof(statline),in);

		if( strcmp(type,"head") == 0 ){
			head = RFC822_readHeader(in,1);
			body = 0;
		}else
		if( strcmp(type,"mime") == 0 ){
			head = RFC822_readHeader(in,1);
			body = 1;
		}else
		if( strcmp(type,"body") == 0 ){
			head = NOHEAD;
			body = 1;
		}else{
sv1log("CFI/MSG ---- unknown type: %s\n",type);
			exit(1);
		}
		eoh = strSeekEOH(head);
		if( *eoh == '.' ){
			strcpy(eohb,eoh);
			*eoh = 0;
		}else	eohb[0] = 0;

		endline[0] = 0;
		fseek(tmp,0,0);
		if( body ){
			relayRESPBODY(in,tmp,endline,sizeof(endline));
			fflush(tmp);
			Ftruncate(tmp,0,2);
			fseek(tmp,0,0);
		}else{
			Ftruncate(tmp,0,0);
		}
		if( http_conv(tmp,out,conninfo,convspec,clproto,statline,head,body) == 0 ){
			fputs(statline,out);
			fputs(head,out);
			simple_relayf(tmp,out);
		}
		fputs(eohb,out);
		fputs(endline,out);

		if( head != NOHEAD )
			free(head);
	}
}
