/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1995 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, modify, and distribute this material for any
purpose and without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies, and
that the name of ETL not be used in advertising or publicity pertaining
to this material without the specific, prior written permission of an
authorized representative of ETL.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	rfc822.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:
History:
	941008	extracted from nntp.c
	950312	encode/decode parts in a multipart message
	951029	extracted from mime.c of DeleGate
//////////////////////////////////////////////////////////////////////#*/
#include "mime.h"
char *RFC822_fgetsHeaderField();
char *nextField();

extern char *getFieldValue2();
#define getFieldValue(str,fld,buf,siz) getFieldValue2(str,fld,buf,siz)

char *findField(head,field,value)
	char *head,*field,**value;
{	char *nsp,*csp,*fsp,*fdp;
	int flen;

	if( field == NULL || field[0] == 0 )
		flen = 0;
	else	flen = strlen(field);

	for( csp = head; *csp; csp = nsp ){
		if( flen == 0 )
			while( *csp && *csp != ':' && *csp != ' ' )
				csp++;

		if( flen == 0 || strncasecmp(csp,field,flen) == 0 )
		if( csp[flen] == ':' || csp[flen] == ' ' ){
			if( flen == 0 )
				fsp = head;
			else{
				fsp = csp;
				csp += flen;
			}
			while( *csp == ' ' ) csp++;
			if( *csp == ':' ) csp++;
			while( *csp == ' ' ) csp++;
			if( value )
				*value = csp;
			return fsp;
		}
		if( nsp = strchr(csp,'\n') )
			nsp++;
		else	return 0;
	}
	return 0;
}

rmField(head,field)
	char *head,*field;
{	char *fp,*np;
	int nf;

	nf = 0;
	while( fp = findField(head,field,NULL) ){
		if( np = strchr(fp,'\n') )
			strcpy(fp,np+1);
		else	*fp = 0;
		nf++;
	}
	return nf;
}

char *findFieldValue(head,field)
	char *head,*field;
{	char *value;

	if( findField(head,field,&value) )
		return value;
	else	return NULL;
}

char *RFC822_valuescan(vp,value,size)
	char *vp,*value;
{	char ch,*bp;
	int cc;
	int pch = -1;

	bp = value;
	for( cc = 1; cc < size && (ch = *vp++); cc++ ){
		if( ch == '\r' )
			continue;

		if( ch == '\n' ){
			if( *vp == ' ' || *vp == '\t' )
				continue;
			break;
		}
		if( ch == '\t' )
			ch = ' ';

		if( ch == ' ' && pch == ' ' )
			continue;
		pch = ch;

		*bp++ = ch;
	}
	*bp = 0;
	return vp;
}

char *getFieldValue2(head,field,value,size)
	char *head,*field,*value;
{	char *vp;

	if( vp = findFieldValue(head,field) ){
		RFC822_valuescan(vp,value,size);
		return value;
	}
	value[0] = 0;
	return NULL;
}

char *RFC822_decompField2(head,fname,value,size)
	char *head,*fname,*value;
{	char *hp,hc,*dp;

	if( fname != NULL ){
		dp = fname;
		for( hp = head; hc = *hp; hp++ ){
			if( hc == ' ' || hc == ':' ){
				hp++;
				break;
			}
			*dp++ = hc;
		}
		*dp = 0;
		RFC822_valuescan(hp,value,size);
	}else
	if( value != NULL )
		getFieldValue(head,"",value,size);
}

char *fgetsHeaderField(hfp,name,value,size)
	FILE *hfp;
	char *name,*value;
{	char line[1024],*vp;
	int len;
	int found;
	int off;

	len = strlen(name);
	found = 0;
	value[0] = 0;

	off = ftell(hfp);
	for(;;){
		if( RFC822_fgetsHeaderField(line,sizeof(line),hfp) == NULL )
			break;
		if( line[0] == '\r' || line[0] == '\n' )
			break;
		if( strncasecmp(line,name,len) == 0 ){
			vp = line + len;
			while( *vp == ' ' )
				vp++;
			if( *vp == ':' ){
				found++;
				vp++;
				RFC822_strip_lwsp(vp,value,size);
			}
		}
	}
	fseek(hfp,off,0);

	if( found )
		return value;
	else	return NULL;
}

static strsubstr(n1,n2)
	char *n1,*n2;
{
	return strstr(n1,n2) == n1 || strstr(n2,n1) == n2;
}
int (*NNTP_nodematch)() = strsubstr;

findXref(afp,matchfunc,host,xref,size)
	FILE *afp;
	int (*matchfunc)();
	char *host,*xref;
{	char line[1024],node[256];
	char *dp,dc,*ep;
	int off,found;

	if( matchfunc == NULL )
		matchfunc = strsubstr;

	xref[0] = 0;
	off = ftell(afp);
	found = 0;

	while( fgets(line,sizeof(line),afp) != NULL ){
		if( line[0] == '\r' || line[0] == '\n' )
			break;
		if( line[0] == '.' )
		if( line[1] == '\r' || line[1] == '\n' )
			break;

		if( isspace(line[0]) )
			continue;

		if( (dp = strchr(line,':')) == 0 )
			continue;
		*dp++ = 0;
		if( strcasecmp(line,"Xref") != 0 )
			continue;

		while( *dp && isspace(*dp) )
			dp++;

		wordscanX(dp,node,sizeof(node));
		if( !(*matchfunc)(host,node) ) 
			continue;

		if( ep = strpbrk(dp,"\r\n") )
			*ep = 0; /* must treat folded line ... */

		strncpy(xref,dp,size-1);
		xref[size-1] = 0;
		found = 1;
		break;
	}
	fseek(afp,off,0);
	return found;
}

selectXref(host,xref1,xref2)
	char *host,*xref1,*xref2;
{	char *sp,*dp,*xp,name[512],host1[512],ch;
	int getit;
	int len;

	len = strlen(host);
	sp = xref1;
	dp = xref2;
	getit = 0;

	while( ch = *sp ){
		if( ch != ' ' && ch != '\t' ){
			getit = 0;
			if( strncasecmp(sp,"Xref",4) == 0 )
			if( xp = strchr(sp,':') ){
				wordscanX(xp+1,host1,sizeof(host1));

				if( strncasecmp(host1,host,len) == 0 )
				if( host1[len] == 0 || host1[len] == '.' )
					getit = 1;
			}
		}
		while( ch = *sp ){
			sp++;
			if( getit )
				*dp++ = ch;
			if( ch == '\n' )
				break;
		}
	}
	*dp = 0;
}

static char *separator(field)
	char *field;
{
	if( strcasecmp(field,"To")==0 || strcasecmp(field,"Cc")==0 )
		return ", ";
	if( strcasecmp(field,"Received") == 0 )
		return "; ";
	return " ";
}

RFC821_skipheader(afp,out,field)
	FILE *afp,*out;
	char *field;
{	char line[1024],*lp,*dp,selected[1024];
	int flen,rcode,nputs;

	rcode = EOF;
	selected[0] = 0;
	nputs = 0;
	if( field != NULL )
		flen = strlen(field);
	else	flen = 0;

	while( fgets(line,sizeof(line),afp) != NULL ){
		if( line[0] == '\r' || line[0] == '\n' ){
			rcode = 0;
			break;
		}
		if( line[0] == '.' )
		if( line[1] == '\r' || line[1] == '\n' )
			break;

		if( out != NULL ){
		    if( field == NULL )
			fputs(line,out);
		    else{
			if( selected[0] == 0 ){
				if( line[flen] != ':' )
					continue;
				if( strncasecmp(line,field,flen) != 0 )
					continue;
				for( lp = line+flen+1; *lp; lp++ )
					if( !isspace(*lp) )
						break;
				if( nputs++ != 0 )
					fputs(separator(field),out);
				if( dp = strpbrk(line,"\r\n\f") )
					*dp = 0;
				fputs(lp,out);
			}else{
				if( isspace(line[0]) ){
					fputs(" ",out);
					fputs(line+1,out);
				}else	selected[0] = 0;
			}
		    }
		}
	}
	return rcode;
}
RFC821_skipbody(afp,out,line,size)
	FILE *afp,*out;
	char *line;
{	char linebuf[1024];
	int rcc;

	if( line == NULL ){
		line = linebuf;
		size = sizeof(linebuf);
	}
	*line = 0;
	rcc = 0;
	for(;;){
		if( fgets(line,size,afp) == NULL ){
			line[0] = 0;
			break;
		}
		if( line[0] == '.' )
		if( line[1] == '\r' || line[1] == '\n' )
			break;

		rcc += strlen(line);
		if( out != NULL )
			fputs(line,out);
	}
	return rcc;
}

char *RFC822_readHeader(in,seeEOR)
	FILE *in;
{	char *lp,buff[0x10000];

	lp = buff;
	for(;;){
		if( fgets(lp,1024,in) == NULL )
			break;
		if( lp[0] == '\r' || lp[0] == '\n' )
			break;

		if( seeEOR )
		if( lp[0] == '.' && (lp[1] == '\r' || lp[1] == '\n') )
			break;

		lp += strlen(lp);
	}
	return stralloc(buff);
}

static charset_namelen(name)
	char *name;
{	char *np;
	int nc;

	for( np = name; nc = *np; np++ )
		if( nc != '_' )
		if( nc != '-' && !isalnum(nc) )
			break;
	return np - name;
}

myMIMEversion(ver)
	char *ver;
{
	sprintf(ver,"1.0 (generated by MimeKit/%s)",MimeKit_Version);
}

char *strSeekEOH(head)
	char *head;
{	char *hp,ch;
	int top;

	top = 1;
	for( hp = head; ch = *hp; hp++ ){
		if( top ){
			if( ch == '\r' || ch == '\n' )
				break;
			if( ch == '.' && (hp[1] == '\r' || hp[1] == '\n') )
				break;
		}
		top = (ch == '\n');
	}
	return hp;
}
insert_ctype(head,ctype)
	char *head,*ctype;
{	char *ctp;
	char EOL[4],ver[128],tmp[0x10000];

	tmp[0] = 0;
	if( strstr(head,"\r\n") )
		strcpy(EOL,"\r\n");
	else	strcpy(EOL,"\n");

	ctp = strSeekEOH(head);
	strcpy(tmp,ctp);

	if( findFieldValue(head,"MIME-Version") == NULL ){
		myMIMEversion(ver);
		sprintf(ctp,"MIME-Version: %s%s",ver,EOL);
		ctp += strlen(ctp);
	}
	sprintf(ctp,"Content-Type: %s%s",ctype,EOL);
	strcat(ctp,tmp);
}

get_charset(ctype,chset,size)
	char *ctype,*chset;
{	char *csp;
	int len;

	if( csp = strcasestr(ctype,"charset=") ){
		csp += strlen("charset=");
		if( *csp == '"' )
			csp++;
		len = charset_namelen(csp);
		if( size < len+1 )
			Strncpy(chset,csp,size);
		else	Strncpy(chset,csp,len+1);
		return len;
	}
	return 0;
}

replace_charset(head,charset)
	char *head,*charset;
{	char *ctp;

	if( charset == NULL )
		return 0;

	if( (ctp = findFieldValue(head,"Content-Type")) == NULL ){
		char ctype[1024];
		sprintf(ctype,"text/plain; charset=%s",charset);
		insert_ctype(head,ctype);
		return 0;
	}
	return replace_charset_value(ctp,charset,0);
}

replace_charset_value(ctype,charset,force)
	char *ctype,*charset;
{	char *csp,*cst;
	char tmp[0x10000];

	if( csp = strcasestr(ctype,"charset=") ){
		csp += strlen("charset=");
		if(*csp == '"')
			csp++;
		if( strncasecmp(csp,charset,strlen(charset)) == 0 )
			return 0;
		cst = csp + charset_namelen(csp);
		strcpy(tmp,cst);
		strcpy(csp,charset);
		strcat(csp,tmp);
	}else{
		if( (cst = strpbrk(ctype,";\r\n")) == 0 ){
			if( force )
				cst = ctype + strlen(ctype);
			else	return 0;
		}
		strcpy(tmp,cst);
		sprintf(cst,"; charset=%s",charset);
		strcat(cst,tmp);
	}
	return 1;
}

erase_charset_param(ctype,charset)
	char *ctype,*charset;
{	char *sp,*np,*pat,*cset,csetb[32];
	int len;

	if( charset )
		cset = charset;
	else	cset = csetb;

	pat = "charset=";
	len = strlen(pat);
	for( sp = ctype; *sp; sp++ )
	if( *sp == ';' ){
		while( isspace(sp[1]) )
			sp++;
		if( strncasecmp(sp+1,pat,len) == 0 ){
			np = valuescanX(sp+1+len,cset,sizeof(csetb));
			strcpy(sp,np);
			return np - sp;
		}
	}
	return 0;
}

replaceFieldValue(head,field,value)
	char *head,*field,*value;
{	char *esp;
	char tmp[0x10000];
	char ch,*np;

	if( (esp = findFieldValue(head,field)) == NULL ){
		np = strSeekEOH(head);
		strcpy(tmp,np);
		sprintf(np,"%s: %s\r\n%s",field,value,tmp);
		return 0;
	}

	if( strncasecmp(esp,value,strlen(value)) == 0 )
	{
		lineScan(esp,tmp);
		if( strcasecmp(tmp,value) == 0 )
		return 0;
	}

	/*
	for( np = esp; ch = *np; np++ )
		if( ch == '\r' || ch == '\n' )
			break;
	*/
	np = nextField(esp,0);
	if( head < np && np[-1] == '\n' ){
		np--;
		if( head < np && np[-1] == '\r' )
			np--;
	}
	strcpy(tmp,np);
	strcpy(esp,value);
	strcat(esp,tmp);
	return 1;
}

RFC822_addHeaderField(dst,src)
	char *dst,*src;
{	char *tp;

	if( (tp = strstr(dst,"\r\n\r\n")) && tp[4] == 0  ){
		strcpy(tp+2,src);
		strcat(tp,"\r\n");
	}else{
		Strins(dst,src);
	}
}

replaceContentType(head,type)
	char *head,*type;
{	char *ctp,*cst;
	char tmp[0x10000],*tp;

	ctp = findFieldValue(head,"Content-Type");
	if( ctp == NULL ){
		tp = stralloc(head);
		sprintf(head,"Content-Type: %s\r\n%s",type,tp);
		free(tp);
		return 0;
	}
	if( (cst = strpbrk(ctp,";\r\n")) == 0 )
		return -1;
	strcpy(tmp,cst);
	strcpy(ctp,type);
	strcat(ctp,tmp);
	return 1;
}

char *RFC822_fgetsHeaderField(line,size,fp)
	char *line;
	FILE *fp;
{	char *lp;
	int rcc,rem,ch;
	
	if( fgets(line,size,fp) == NULL )
		return NULL;
	lp = line;
	if( *lp == 0 || *lp == '\r' || *lp == '\n' || isEOR(lp) )
		return line;

	rcc = strlen(line);
	lp = line + rcc;
	rem = size - rcc;

	while( 80 < rem ){
		ch = getc(fp);
		if( ch == EOF )
			break;
		if( ch != ' ' && ch != '\t' && ch != '#' ){
			ungetc(ch,fp);
			break;
		}
		*lp++ = ch; *lp = 0; rem--;
		if( fgets(lp,rem,fp) == NULL )
			break;
		rcc = strlen(lp);
		lp += rcc;
		rem -= rcc;
	}
	if( lp == line )
		return NULL;
	else	return line;
}

char *nextField(field,ignEOH)
	char *field;
{	char *np,nc;
	int top;

	top = 0;
	for( np = field; nc = *np; np++ ){
		if( top ){
			if( nc == ' ' || nc == '\t' )
				top = 0;
			else
			if( !ignEOH || (nc != '\n' && nc != '\r') )
				break;
		}else{
			if( nc == '\n' )
				top = 1;
			else	top = 0;
		}
	}
	return np;
}

filterFields(spec,head)
	char *spec,*head;
{	char *sp;
	char field[1024],aval[1024];
	char *hp,*bp,*tp;
	int pass,match;
	int plen,flen;
	char *headbuf,*hbp;

	pass = 0;
	sp = spec;

	for( sp = spec; *sp; sp = nextField(sp,1) ){
		if( strncasecmp(sp,"Pass/",5) == 0 ){
			pass = 1;
			headbuf = stralloc(head);
			hbp = headbuf;
			break;
		}
	}
	if( pass )
		plen = 5;
	else	plen = 7;

	for( sp = spec; *sp; sp = nextField(sp,1) ){
		if( pass ){
			if( strncasecmp(sp,"Pass/",5) != 0 )
				continue;
		}else{
			if( strncasecmp(sp,"Remove/",7) != 0 )
				continue;
		}

		fieldScan(sp+plen,field,aval);
		flen = strlen(field);

		for( hp = head; *hp; ){
			match = 0;
			if( strncasecmp(hp,field,flen) == 0 )
			if( hp[flen] == ':' )
			if( strstr(&hp[flen+1],aval) )
				match = 1;

			tp = nextField(hp,0);
			if( pass ){
				if( match ){
					for( bp = hp; *bp && bp != tp; bp++ )
						*hbp++ = *bp;
				}
				hp = tp;
			}else{
				if( match )
					strcpy(hp,tp);
				else	hp = tp;
			}

			if( *hp == '\r' || *hp == '\n'
			 || *hp == '.' && (hp[1] == '\r' || hp[1] == '\n') )
				break;
		}
	}
	if( pass ){
		*hbp = 0;
		strcpy(hbp,strSeekEOH(head));
		strcpy(head,headbuf);
		free(headbuf);
	}
}

removeFields(head,field,wild)
	char *head,*field;
{	char *dp,*sp,ch;
	int flen,top,skip;

	flen = strlen(field);
	top = 1;
	skip = 0;
	dp = head;
	for( sp = head; ch = *sp; sp++ ){
		if( top ){
			if( strncmp(sp,field,flen) == 0
			 && (wild || (sp[flen]==':'||sp[flen]==' ')) )
				skip = 1;
			else
			if( ch != ' ' && ch != '\t' )
				skip = 0;
		}
		if( !skip )
			*dp++ = ch;

		if( ch == '\n' )
			top = 1;
		else	top = 0;
	}
	*dp = 0;
}

/*
 *    MATCHFIELDS
 *	Returns the first "field" which match with the "spec":
 *	  [Tail/] field [*] : [ivalue]
 */
char *matchFields(spec,field,ivalue)
	char *spec,*field,*ivalue;
{	char *afield,*avalue,aval[1024];
	int tail = 0;

	afield = findField(spec,field,&avalue);
	if( afield == NULL )
		return "";
	do{
		if( tail = strncmp(afield,"Tail/",5) == 0 )
			afield += 5;

		getFieldValue(afield,field,aval,sizeof(aval));
		if( tail ){
			if( strtailstr(ivalue,aval) != NULL )
				return afield;
		}else{
			if( strstr(ivalue,aval) != NULL )
				return afield;
		}
	} while( afield = findField(afield+1,field,&avalue) );
	return NULL;
}

relayRESPBODY(fs,tc,line,size)
	FILE *fs,*tc;
	char *line;
{
	for(;;){
		line[0] = 0;
		if( fgets(line,size,fs) == NULL )
			break;
		if( isEOR(line) )
			break;
		fputs(line,tc);
	}
}

RFC822_strip_lwsp(src,dst,size)
	char *src,*dst;
{	char *sp,*dp,sc,pc,nc;
	char *dx;
	int nonsp;

	nonsp = 0;
	pc = 0;
	dp = dst;
	dx = dst + size - 1;
	for( sp = src; sc = *sp; sp++ ){
		if( dx <= dp )
			break;
		if( sc == '\r' || sc == '\n' ){
			nonsp = 0;
		}else
		if( sc == '\t' || sc == ' ' ){
			if( dp == dst )
				continue;
			nc = sp[1];
			if( nc == '\r' || nc == '\n' || nc == '\0' )
				continue;
			if( pc != ' ' )
				*dp++ = pc = ' ';
		}else{
			nonsp++;
			*dp++ = pc = sc;
		}
	}
	*dp = 0;
}

RFC822_strip_commentX(in,out,siz)
	char *in,*out;
{	char *sp,*dp,*xp,sc;
	int lev;

	lev = 0;
	dp = out;
	xp = out + siz - 1;

	for( sp = in; sc = *sp; sp++ ){
		if( xp <= dp )
			break;
		if( sc == '(' )
			lev++;
		else
		if( sc == ')' ){
			lev--;
			if( lev == 0 && (sp[1] == ' ' || sp[1] == '\t') )
				sp++;
		}else
		if( lev == 0 ){
			*dp++ = sc;
		}
	}
	*dp = 0;
}

/*
 * any non-ASCII (7bit x multi-bytes code) strings must be ignored...
 * but such strings are expected to be encoded in MIME...
 */
RFC822_addresspartX(in,out,siz)
	char *in,*out;
{	char buff[2048],*sp,*dp,*xp;
	char inx[2048];
	unsigned char sc;

	MIME_strHeaderEncode(in,inx,sizeof(inx)); /* non-ASCII escaped */
	RFC822_strip_commentX(inx,buff,sizeof(buff));
	dp = out;
	xp = out + siz - 1;

	for( sp = buff; sc = *sp; sp++ ){
		if( xp <= dp )
			break;
		if( sc == '<' ){
			dp = out;
			for( sp += 1; sc = *sp; sp++ ){
				if( isspace(sc) )
					continue;
				if( sc == '>' )
					break;
				*dp++ = sc;
			}
			break;
		}else
		if( isspace(sc) )
			continue;
		else{
			*dp++ = sc; 
		}
	}
EXIT:
	*dp = 0;
}


msg_charcount(fp,chcount)
	FILE *fp;
	int chcount[];
{	int off;
	char line[1024],xline[1024];
/* bad for BCGCC
	int ccx[64],oc,ci,ch;
*/
	char ccx[64];
	int oc,ci,ch;
	int uu;

	off = ftell(fp);
	CCXcreate("*","E",ccx);
	for( ci = 0; ci < 256; ci++ )
		chcount[ci] = 0;

	/* skip the header if necessary... */

	uu = 0;
	while( fgets(line,sizeof(line),fp) ){
		if( uu_skip(&uu,line) )
			continue;

		oc = CCXexec(ccx,line,strlen(line),xline,sizeof(xline));
		for( ci = 0; ci < oc; ci++ ){
			ch = 0xFF & xline[ci];
			if( ch == '!' ){
				/* count only '!' at the end of sentense ... */
				if( ci == 0 ) /* maybe "diff" ountput ... */
					continue;
				if( strchr("!? <\t\r\n",xline[ci+1]) == 0 )
					continue;
			}
			chcount[ch] += 1;
		}
	}
	fseek(fp,off,0);
}

mapEmailAddr(spec,addr,xaddr,xsiz)
	char *spec,*addr,*xaddr;
{	char *dp,*xp,*fp,fc,*op;
	char oloc[256];
	char ohost[256],odomL[256];
	char odom1[256],odomU[256],odomT[256];
	char ogdom[256],odfull[256];
	int xrem,os;

	dp = wordscanY(addr,oloc,sizeof(oloc),"^@");
	ohost[0] = odomL[0] = 0;
	odom1[0] = odomU[0] = odomT[0] = 0;
	ogdom[0] = odfull[0] = 0;
	if( *dp == '@' ){
		wordScan(dp+1,odfull);
		wordScan(dp+1,ohost);
		wordScan(dp+1,ogdom);
		generic_domain(ogdom); /* registered domain part */
		if( dp = strstr(ohost,ogdom) )
			*dp = 0;
		if( dp = strrchr(ohost,'.') ){
			*dp = 0;
			wordScan(dp+1,odomL);
		}
		if( dp = strchr(ogdom,'.') )
			wordScan(dp+1,odomU);
		if( dp = strrchr(ogdom,'.') )
			wordScan(dp+1,odomT);
		wordscanY(ogdom,odom1,sizeof(odom1),"^.");
	}
	/*
	syslog_ERROR("### [%s] @ [%s] [%s] [%s] [%s]\n",
		oloc,ohost,odomL,odom1,ogdom);
	*/

	xp = xaddr;
	xrem = xsiz;
	for( fp = spec; fc = *fp; fp++ ){
		op = 0;
		if( fc != '%' ){
			op = fp;
			os = 1;
		}else{
			if( (fc = *++fp) == 0 )
				break;

			/* l@h.L.r.U.c */
			switch( fc ){
			  case '%': op - "%"; break;
			  case 'l': op = oloc; break; /* l */
			  case 'h': op = ohost; break; /* h */
			  case 'L': op = odomL; break; /* L */
			  case 'r': op = odom1; break; /* r */
			  case 'u': op = odomU; break; /* U.c */
			  case 'c': if( strlen(odomT) == 2 ) op = odomT;
				break; /* c */
			  case 't': op = odomT; break; /* c */
			  case 'g': op = ogdom; break; /* r.U.c */
			  case 'f': op = odfull; break; /* h.L.r.U.c */
			}
			if( op )
				os = strlen(op);
		}
		if( op == 0 )
			continue;

		xrem -= os;
		if( xrem < 1 ){
			syslog_ERROR("mapEmailAddr overflow %d: %s[%s]%d\n",
				xsiz,xaddr,op,os);
			break;
		}
		bcopy(op,xp,os);
		xp += os;
	}
	*xp = 0;
	/*
	syslog_ERROR("### -> [%s]\n",xaddr);
	*/
}
rewriteEmailAddr(head,field,spec)
	char *head,*field,*spec;
{	char from[256],addr[256],xaddr[256];
	char *val,*dp;

	val = findFieldValue(head,field);
	if( val == NULL )
		return;
	/*
	lineScan(val,from);
	*/
	RFC822_valuescan(val,from,sizeof(from));
	RFC822_addresspartX(from,addr,sizeof(addr));

	mapEmailAddr(spec,addr,xaddr,sizeof(xaddr));

	/*
	ofrom[0] = 0;
	if( curAnum )
		sprintf(ofrom,"%d.%s",curAnum,curGroup);
	else
	if( getFV(head,"Message-ID",msgid) ){
		val = msgid;
		if( *val == '<' ) val++;
		if( vp = strchr(val,'>') )
			*vp = 0;
		quotedFrom(val,ofrom);
	}
	*/

	if( dp = strstr(from,addr) ){
		if( sizeof(from) <= strlen(from)+strlen(xaddr)-strlen(addr) ){
			syslog_ERROR("rewriteEmailAddr overflow: %s [%s][%s]\n",
				from,addr,xaddr);
			xaddr[strlen(addr)] = 0;
		}
		strsubst(dp,addr,xaddr);
	}else	strcpy(from,xaddr);

	replaceFieldValue(head,field,from);
}

/*
static quotedFrom(from,qfrom)
	char *from,*qfrom;
{	char *fp,*qfp,ch;

	qfp = qfrom;
	for( fp = from; ch = *fp; fp++ ){
		if( IS_GROUPNAMECH(ch) ){
			*qfp++ = ch;
		}else{
			if( ch == '+' ) strcpy(qfp,"++"); else
			if( ch == '@' ) strcpy(qfp,"+."); else
					sprintf(qfp,"+%02x",ch&0xFF);
			qfp += strlen(qfp);
		}
	}
	*qfp = 0;
}
*/

scanAddrInBody(spec,line)
	char *spec,*line;
{	char *top,*atp,*endp,*bgnp,*dp,ch,nch;
	char addr[256],xaddr[256];
	int rew,dot,susp,type,next;

	rew = 0;
	for( top = line; ; top = atp+1 ){
		atp = strchr(top,'@');
		if( atp == 0 )
			break;

		dot = 0;
		for( endp = atp+1; ch = *endp; endp++ ){
			if( ch == '.' )
				dot++;
			else
			if( isalnum(ch) || ch == '-' || ch == '_' ){
			}else{
				break;
			}
		}
		if( dot == 0 ){
			continue;
		}
		switch( ch ){
			case '>': type = 1; break;
			case ')': type = 2; break;
			case ' ':
				if( endp[1] == '(' ) type = 3; else
				if( isalpha(endp[1])) type = 4; else
					continue;
				break;
			default: continue;
		}

		/*
		sv1log("##TR## %d %s",type,top);
		*/

		susp = 0;
		if( type == 1 ){
			if( endp[1] != ' ' )
				susp++;
		}

		for( bgnp = atp-1; top <= bgnp; bgnp-- )
		switch ( ch = *bgnp ){
		  case '-':
		  case '.':
		  case '_':
			susp += 1;
			break;
		  case '+':
			susp += 2;
			break;
		  case '$':
			susp += 10;
			break;
		  case '%':
			break;
		  default:
			if( ch=='<' || isspace(ch) || ch=='(' )
				goto EOA;

			if( isalpha(ch) ){
				nch = bgnp[1];
				if( islower(ch) && isupper(nch) )
					susp += 2;
			}else
			if( isdigit(ch) ){
				nch = bgnp[1];
				susp += 1;
				if( !isdigit(nch) && strchr("@-",nch)==0 )
					susp += 2;
			}else{
				goto NEXT;
			}
		} EOA:

		switch( type ){
			case 1: if( ch != '<' ) goto NEXT; break;
			case 2: if( ch != '(' ) goto NEXT; break;
			case 3: break;
			case 4: if( ch != ' ' ) goto NEXT; break;
		}

		if( type == 1 ){ /* could be <Message-Id> */
			if( (dp = strstr(top,"in <"))
			 || (dp = strstr(top,"message <"))
			 || (dp = strstr(top,"article <"))
			){
				if( dp == top || isspace(dp[-1]) )
					susp += 10;
			}
			if( bgnp == top )
				susp += 3;

			if( 5 <= susp ){
				goto NEXT;
			}
		}
		if( type == 4 ){
			if( strstr(endp," writes") != endp
			 && strstr(endp," wrote") != endp
			)
				goto NEXT;
			/*
			sv1log("##TR4#%sn",endp);
			*/
		}

		linescanX(bgnp+1,addr,endp-bgnp);
		mapEmailAddr(spec,addr,xaddr,sizeof(xaddr));
		/*
		sv1log("##TR## [%s][%s]\n",addr,xaddr);
		*/
		next = strchr(atp+1,'@') != 0;
		strsubst(top,addr,xaddr);
		rew++;
		if( !next )
			break;
		atp = strchr(top,'@');
		if( atp == NULL )
			break;
NEXT:;
	}
	return rew;
}

extern char *MIME_transAddrSpec;
static rewaddr1(where,head,rewaddr)
	char *where,*head,*rewaddr;
{
	if( strcaseeq(where,"Body") )
		MIME_transAddrSpec = rewaddr;
	else	rewriteEmailAddr(head,where,rewaddr);
	return 0;
}
MIME_rewaddrs(rewaddr,head)
	char *rewaddr,*head;
{	char *fmtp,wheb[256];

	if( rewaddr == 0 || *rewaddr == 0 ){
		MIME_transAddrSpec = 0;
		return;
	}
	/*
	sv1log("##TR## transAddrSpec[%s]\n",rewaddr);
	*/
	fmtp = wordscanY(rewaddr,wheb,sizeof(wheb),"^:");
	if( *fmtp != ':' ){
		return;
	}
	strtolower(wheb,wheb);
	scan_commaListL(wheb,0,rewaddr1,head,fmtp+1);
}


extractParam(head,fname,pname,pvalue,pvsize,del)
	char *head,*fname,*pname,*pvalue;
{	char *field1,*fn,*fv,*pp,*dp,*ep,*sp,buf[1024];
	int plen,len;

	plen = strlen(pname);
	for( field1 = head; *field1; ){
		fn = findField(field1,fname,&fv);
		if( fn == NULL )
			break;
		field1 = nextField(fn,0);
		RFC822_valuescan(fv,buf,sizeof(buf));

		pp = strstr(buf,pname);
		if( pp && pp[plen] == '=' ){
			dp = pp + plen + 1;
			if( *dp == '"' ){
				ep = wordscanY(dp+1,pvalue,pvsize,"^\"");
				if( *ep == '"' ) ep++;
			}else	ep = wordscanY(dp,pvalue,pvsize,"^;\r\n");
			if( del ){
				if( *ep == ';' ) ep++;
				if( *ep == ' ' ) ep++;
				ovstrcpy(pp,ep);
				for( sp = buf; *sp; sp++ ){
					if( !isspace(*sp) )
						break;
				}
				if( *sp == 0 ){
					/* become empty */
					ovstrcpy(fn,field1);
				}else{
					len = ep - pp;
					pp = strstr(fv,pname);
					ovstrcpy(pp,pp+len);
				}
			}
			return 1;
		}
	}
	return 0;
}
