/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1994-1999 Yutaka Sato
Copyright (c) 1994-1999 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, and distribute this material for any purpose
and without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	html.c (HTML entity encoder/decoder)
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:

	HTMLCONV=uri:{convTypeList}:{AttrList/EntityList}
	URICONV=convTypeList:AttrList/EntityList
	convType = {full,mount,normal,partial}

	EXAMPLE:
		URICONV="mount,normal,partial:+"
		URICONV="full:+,-HREF/BASE"

History:
	941002	extracted from http.c
	990331	extracted attrWithURL() and made it custmizable
//////////////////////////////////////////////////////////////////////#*/
#include "ystring.h"
/*
#define isAlpha(ch)	(isalpha(ch) || ch == '-')
*/
#define isAlpha(ch)	(isalpha(ch) || ch == '-' || ch == '!' || ch == '#' )

int DECODE_HTML_ENTITIES = 1;
int ENCODE_HTML_ENTITIES = 0;
int PLAIN_TO_HTML_PRE = 0;
int HTMLCONV_DEBUG = 0;
int URL_SEARCH;

static hconv1(conv)
	char *conv;
{	int onoff;

	if( *conv == '!' || *conv == '-' ){
		conv++;
		onoff = 0;
	}else	onoff = 1;

	if( streq(conv,"fullurl") ){
		setURIconv("mount:+");
		setURIconv("full:+,-HREF/BASE");
		return 0;
	}else
	if( streq(conv,"partial") ){
		setURIconv("partial");
		return 0;
	}else
	if( streq(conv,"normal") ){
		setURIconv("normal");
		return 0;
	}

	if( streq(conv,"debug") )  HTMLCONV_DEBUG = onoff; else
	if( streq(conv,"enent") )  ENCODE_HTML_ENTITIES = onoff; else
	if( streq(conv,"deent") )  DECODE_HTML_ENTITIES = onoff; else
	if( streq(conv,"pre")   )  PLAIN_TO_HTML_PRE    = onoff; else
	if( streq(conv,"dump")  )  dumpHTMLCONV(); else
	if( strncmp(conv,"defelem:",8) == 0 ) addTAGelem(conv+8); else
	if( strncmp(conv,"defattr:",8) == 0 ) addTAGattr(conv+8); else
	if( strncmp(conv,"uri:",4) == 0 ) setURIconv(conv+4); else
	if( strncmp(conv,"killtag:",8) == 0 ) setKillTags(conv+8); else
	syslog_ERROR("Error: unknown HTML conversion spec: %s\n",conv);
	return 0;
}
scan_URICONV(_,conv)
	void *_;
	char *conv;
{
	if( streq(conv,"where:any") ) URL_SEARCH = /*ANY*/0xFFFF; else
	if( streq(conv,"dump") )	dumpHTMLCONV(); else
	if( streq(conv,"+") )		setURICONVdefault(0); else
	if( strneq(conv,"defelem:",8) )	addTAGelem(conv+8); else
	if( strneq(conv,"defattr:",8) )	addTAGattr(conv+8); else
					setURIconv(conv);
	return 0;
}
scan_HTMLCONV(_,clist)
	void *_;
	char *clist;
{
	ENCODE_HTML_ENTITIES = 0;
	DECODE_HTML_ENTITIES = 0;
	scan_commaList(clist,0,hconv1);
	return 0;
}

typedef struct {
	char	*_sym;
	int	 _id;
} SymID;
static addsym1(sym,tab)
	char *sym;
	SymID *tab;
{	int ai,id1,maxid,newid;
	char *sym1,symb[128],*sp,sc;

	strcpy(symb,sym);
	sym = symb;
	for( sp = sym; sc = *sp; sp++ )
		if( islower(sc) )
			*sp = toupper(sc);

	maxid = 0;
	for( ai = 0; sym1 = tab[ai]._sym; ai++ ){
		if( strcasecmp(sym1,sym) == 0 )
			return;
		id1 = tab[ai]._id;
		if( (id1 & 0xF) != 0xF && maxid < id1 )
			maxid = id1;
	}
	if( maxid != 0 )
		newid = maxid << 1;
	else	newid = 1;
	tab[ai]._sym = stralloc(sym);
	tab[ai]._id = newid;
	tab[ai+1]._sym = 0;
	syslog_ERROR("HTMLCONV ADDSYMID [%d][%s][%x]\n",ai,sym,tab[ai]._id);
	return 0;
}
static addsyms(tab,syms)
	SymID *tab;
	char *syms;
{
	if( syms[0] == '+' && syms[1] == ',' )
		syms += 2;
	else	tab[2]._sym = 0; /* [0] for "*" and [1] for "+" */
	scan_commaListL(syms,0,addsym1,tab);

/* consistency with rules must be checked after addsyms / clear symtab */
}
static sym2id(tab,sym)
	SymID *tab;
	char *sym;
{	int ai;
	char *sym1;

	for( ai = 0; sym1 = tab[ai]._sym; ai++ )
		if( strcaseeq(sym,sym1) )
			return tab[ai]._id;
	return 0;
}
static char *id2sym0(tab,id)
	SymID *tab;
{	int ai,id1;

	for( ai = 0; id1 = tab[ai]._id; ai++ )
		if( id == id1 )
			return tab[ai]._sym;
	return NULL;
}
static char *id2sym(tab,id)
	SymID *tab;
{	char *sym;

	if( sym = id2sym0(tab,id) )
		return sym;
	else	return "";
}
static sym1(sym,tab,maskp)
	char *sym;
	SymID *tab;
	int *maskp;
{
	*maskp |= sym2id(tab,sym);
	return 0;
}

/* Light weight screening of names to be cared.
 * Possible locations (Nth character in a name) of each character
 * in all of names to be cared are indicated as the Nth bits of
 * each (short) integer for each character.
 */
typedef struct {
	int	n_initDone;
	short	n_chars[256];
	int	n_maxLen;
} NameScreen;
static initNameScreen(tab,nmap)
	SymID *tab;
	NameScreen *nmap;
{	int ai,ci;
	char *ap,ac;

	for( ai = 0; ap = tab[ai]._sym; ai++ ){
		if( isAlpha(*ap) )
		for( ci = 0; ci < 16 && (ac = ap[ci]); ci++ )
			nmap->n_chars[toupper(ac) & 0xFF] |= (1 << ci);
		if( nmap->n_maxLen < ci )
			nmap->n_maxLen = ci;
	}
	nmap->n_initDone = 1;
}
static scanNameScreen(tab,nmap,astr,asym)
	SymID *tab;
	NameScreen *nmap;
	char *astr,*asym;
{	char *op;
	char uc,ac;
	int ci;
	int maxlen,alen;

	/*
	 *	Setup the table for screening attribute names
	 */
	if( nmap->n_initDone == 0 )
		initNameScreen(tab,nmap);

	/*
	 *	Screening attribute names
	 */
	uc = toupper(astr[0]);
	if( (nmap->n_chars[uc & 0xFF] & 1) == 0 )
		return 0;

	op = asym;
	maxlen = nmap->n_maxLen;
	for( ci = 0; ci < maxlen && (ac = astr[ci]); ci++ ){
		if( !isAlpha(ac) )
			break;
		uc = toupper(ac);
		if( (nmap->n_chars[uc & 0xFF] & (1 << ci)) == 0 )
			return 0;
		*op++ = uc;
	}
	if( op == asym )
		return 0;
	*op = 0;
	alen = op - asym;
	return alen;
}

#define NCONVS 64

/* type of URL conversions */
#define C_MOUNTF	0x0001	/* URLs in forwarding */
#define C_MOUNTR	0x0002	/* URLs in response */
#define C_MOUNTD	0x0004	/* -_- notation */
#define C_MOUNT		0x0007
#define C_NORMAL	0x0010
#define C_PARTIAL	0x0020
#define C_FULL		0x0040
#define C_ANY		0x0077
#define C_DEFAULTS	(C_MOUNT|C_NORMAL|C_PARTIAL)
#define C_DEFAULTA	(C_MOUNT|C_NORMAL)
#define C_REQ		0x0100	/* required */
#define C_SPL		0x0200	/* list separated with space */

#define C_SSI		0x1000
#define C_META		0x2000
#define C_APPLET	0x4000	/* conversion for APPLET */
#define C_JAVA		0x8000	/* conversion for Java {APPLET|OBJECT|EMBED} */
#define C_KILL		0x10000 /* disable tag */

int URICONV_MOUNTF =	C_MOUNTF;
int URICONV_MOUNTR =	C_MOUNTR;
int URICONV_MOUNTD =	C_MOUNTD;
int URICONV_MOUNT =	C_MOUNT;
int URICONV_NORMAL =	C_NORMAL;
int URICONV_PARTIAL =	C_PARTIAL;
int URICONV_FULL =	C_FULL;
int URICONV_ANY	=	C_ANY;
int TAGCONV_SSI =	C_SSI;
int TAGCONV_META =	C_META;
int TAGCONV_APPLET =	C_APPLET;
int TAGCONV_JAVA =	C_JAVA;
int TAGCONV_KILL =	C_KILL;

int URICONV_nMOUNTF;
int URICONV_nMOUNTR;
int URICONV_nMOUNTD;
int URICONV_nMOUNT;
int URICONV_nNORMAL;
int URICONV_nPARTIAL;
int URICONV_nFULL;
int TAGCONV_nKILL;

static SymID convsym[NCONVS] = {
	"*",		C_ANY,
	"mountF",	C_MOUNTF,
	"mountR",	C_MOUNTR,
	"mountD",	C_MOUNTD,
	"mount",	C_MOUNT,
	"normal",	C_NORMAL,
	"partial",	C_PARTIAL,
	"full",		C_FULL,
	"killtag",	C_KILL,
	"+",		C_DEFAULTS,

	"required",	C_REQ,
	"splist",	C_SPL,

	"ssi",		C_SSI,
	"meta",		C_META,
	"applet",	C_APPLET,
	"java",		C_JAVA,
	0,
};

/* tag's element names */
#define E_UNKNOWN	0x80000000
#define E_ANY		0xFFFFFFFF
#define E_A		0x00000001
#define E_APPLET	0x00000002
#define E_AREA		0x00000004
#define E_BASE		0x00000008
#define E_BLOCKQUOTE	0x00000010
#define E_BODY		0x00000020
#define E_DEL		0x00000040
#define E_FORM		0x00000080
#define E_FRAME		0x00000100
#define E_HEAD		0x00000200
#define E_IFRAME	0x00000400
#define E_IMG		0x00000800
#define E_INPUT		0x00001000
#define E_INS		0x00002000
#define E_LINK		0x00004000
#define E_OBJECT	0x00008000
#define E_Q		0x00010000
#define E_SCRIPT	0x00020000
#define E_EMBED		0x00040000

#define E_Header	0x00080000
#define E_X_META	0x00100000
/* HTTP header and META HTTP-EQUIV should be identified as a context ? */

#define E_DEFINEDS	0x001FFFFF

#define E_X_ECHO	0x01000000
#define E_X_INCLUDE	0x02000000
#define E_X_FSIZE	0x04000000
#define E_X_FLASTMOD	0x08000000
#define E_X_EXEC	0x10000000
#define E_X_CONFIG	0x20000000
#define E_X_SSITAGS	0x3F000000

static SymID elemsym[NCONVS] = {
	"*",		E_ANY,
	"+",		E_DEFINEDS,
	"A",		E_A,
	"APPLET",	E_APPLET,
	"AREA",		E_AREA,
	"BASE",		E_BASE,
	"BLOCKQUOTE",	E_BLOCKQUOTE,
	"BODY",		E_BODY,
	"DEL",		E_DEL,
	"EMBED",	E_EMBED,
	"FORM",		E_FORM,
	"FRAME",	E_FRAME,
	"HEAD",		E_HEAD,
	"IFRAME",	E_IFRAME,
	"IMG",		E_IMG,
	"INPUT",	E_INPUT,
	"INS",		E_INS,
	"LINK",		E_LINK,
	"OBJECT",	E_OBJECT,
	"Q",		E_Q,
	"SCRIPT",	E_SCRIPT,
	"Header",	E_Header,

	"META",		E_X_META,
	"!--#echo",	E_X_ECHO,
	"!--#include",	E_X_INCLUDE,
	"!--#fsize",	E_X_FSIZE,
	"!--#flastmod",	E_X_FLASTMOD,
	"!--#exec",	E_X_EXEC,
	"!--#config",	E_X_CONFIG,
	"ssitags",	E_X_SSITAGS,
	0,
};
static addTAGelem(name) char *name; { addsyms(elemsym,name); }

char TAGCONV_TAGEND[] =	"-"; /* pseudo attribute for end of tag */

/* tag's attribute names */
#define A_ANY		0xFFFFFFFF
#define A_ACTION	0x00000001
#define A_ARCHIVE	0x00000002
#define A_BACKGROUND	0x00000004
#define A_CITE		0x00000008
#define A_CLASSID	0x00000010
#define A_CODEBASE	0x00000020
#define A_DATA		0x00000040
#define A_HREF		0x00000080
#define A_LONGDESC	0x00000100
#define A_PROFILE	0x00000200
#define A_SRC		0x00000400
#define A_USEMAP	0x00000800
#define A_IMAGEMAP	0x00001000
#define A_SCRIPT	0x00002000
#define A_URL		0x00004000
#define A_Location	0x00008000
#define A_ContLocation	0x00010000
#define A_Set_Cookie	0x00020000
#define A_DEFINEDS	0x0003FFFF

#define A_CODE		0x00100000
#define A_END		0x00200000

#define A_X_HTTPEQUIV	0x01000000
#define A_X_VAR		0x02000000
#define A_X_FILE	0x04000000
#define A_X_VIRTUAL	0x08000000
#define A_X_FILES	(A_X_FILE|A_X_VIRTUAL)
#define A_X_CMD		0x10000000
#define A_X_CGI		0x20000000
#define A_X_CMDS	(A_X_CMD|A_X_CGI|A_X_VIRTUAL)
#define A_X_TIMEFMT	0x40000000
#define A_X_SIZEFMT	0x80000000
#define A_X_CONFIGS	(A_X_TIMEFMT|A_X_SIZEFMT)

static SymID attrsym[NCONVS] = {
	"*",		A_ANY,
	"+",		A_DEFINEDS,
	"ACTION",	A_ACTION,
	"ARCHIVE",	A_ARCHIVE,
	"BACKGROUND",	A_BACKGROUND,
	"CITE",		A_CITE,
	"CLASSID",	A_CLASSID,
	"CODE",		A_CODE,
	"CODEBASE",	A_CODEBASE,
	"DATA",		A_DATA,
	"HREF",		A_HREF,
	"LONGDESC",	A_LONGDESC,
	"PROFILE",	A_PROFILE,
	"SRC",		A_SRC,
	"USEMAP",	A_USEMAP,
	"IMAGEMAP",	A_IMAGEMAP,
	"SCRIPT",	A_SCRIPT,
	"URL",		A_URL,
	"Location",	A_Location,
	"Content-Location",A_ContLocation,
	"Set-Cookie",	A_Set_Cookie,
	TAGCONV_TAGEND,	A_END,

	"HTTP-EQUIV",	A_X_HTTPEQUIV,
	"var",		A_X_VAR,
	"file",		A_X_FILE,
	"virtual",	A_X_VIRTUAL,
	"cgi",		A_X_CGI,
	"cmd",		A_X_CMD,
	"timefmt",	A_X_TIMEFMT,
	"sizefmt",	A_X_SIZEFMT,
	0,
};
static addTAGattr(attr) char *attr; { addsyms(attrsym,attr); }


typedef struct {
	int	 a_convtype;
	int	 a_id; /* atrribute */
	int	 a_context; /* tag */
} Attr;

typedef struct {
	NameScreen a_AttrNames;
	Attr	r_Attr[NCONVS*2];
	int	r_AttrInit;
	int	r_AttrX;
	int	a_Serno;
} HtmlConv;
static HtmlConv *htmlConvs;
minit_html()
{
	if( htmlConvs == 0 )
		htmlConvs = NewStruct(HtmlConv);
}
#define htmlConv	htmlConvs[0]
#define attrNames	htmlConv.a_AttrNames
#define	rewriteAttr	htmlConv.r_Attr
#define	rewriteAttrInit	htmlConv.r_AttrInit
#define	rewriteAttrX	htmlConv.r_AttrX
#define attrSerno	htmlConv.a_Serno

static Attr rewriteAttrDefault[NCONVS] = {
{C_DEFAULTS|C_REQ,A_ACTION,	E_FORM			},
{C_DEFAULTS|C_SPL,A_ARCHIVE,	E_OBJECT		},
{C_DEFAULTS,A_BACKGROUND,	E_BODY			},
{C_DEFAULTS,A_CITE,		E_BLOCKQUOTE |E_Q |E_DEL |E_INS},
{C_DEFAULTS,A_CLASSID,		E_OBJECT		},
{C_DEFAULTS,A_CODEBASE,		E_OBJECT |E_APPLET	},
{C_DEFAULTS,A_DATA,		E_OBJECT		},
{C_DEFAULTS,A_HREF,		E_A |E_AREA |E_LINK	},
{C_DEFAULTA,A_HREF,		E_BASE			},
{C_DEFAULTS,A_LONGDESC,		E_IMG |E_FRAME |E_IFRAME},
{C_DEFAULTS,A_PROFILE,		E_HEAD			},
{C_DEFAULTS|C_REQ,A_SRC,	E_IMG},
{C_DEFAULTS,A_SRC,		E_INPUT |E_SCRIPT |E_FRAME |E_IFRAME},
{C_DEFAULTS,A_USEMAP,		E_IMG |E_INPUT |E_OBJECT},

{C_DEFAULTA,A_IMAGEMAP,		E_ANY}, /* obsolete ? */
{C_DEFAULTA,A_SCRIPT,		E_ANY}, /* obsolete ? */
{C_DEFAULTA,A_URL,		E_ANY}, /* obsolete ? */

{C_DEFAULTA,A_ContLocation,	E_Header},
{C_DEFAULTA,A_Location,		E_Header}, /* to be supported ? */
{C_DEFAULTA,A_Set_Cookie,	E_Header}, /* to be supported ? */

{C_META,    A_X_HTTPEQUIV,      E_X_META},
{C_SSI,     A_X_VAR,		E_X_ECHO},
{C_SSI,     A_X_FILES,		E_X_INCLUDE},
{C_SSI,     A_X_FILES,		E_X_FSIZE},
{C_SSI,     A_X_FILES,		E_X_FLASTMOD},
{C_SSI,     A_X_CMDS,		E_X_EXEC},
{C_SSI,     A_X_CONFIGS,	E_X_CONFIG},

{C_JAVA,    A_CODE|A_END,	E_APPLET |E_EMBED |E_OBJECT},
{C_APPLET,  A_CODE|A_END,	E_APPLET},
0
};

static list2mask(tab,list)
	SymID *tab;
	char *list;
{	int mask;

	mask = 0;
	scan_commaList(list,0,sym1,tab,&mask);
	return mask;
}
static mask2list(tab,mask,list)
	SymID *tab;
	char *list;
{	int mi;
	char *bp,*sym;
	int flag;

	bp = list;
	for( mi = 0; mi < 32; mi++ ){
		flag = 1 << mi;
		if( tab == convsym
		&& (flag & C_MOUNT) && (mask & C_MOUNT) == C_MOUNT ){
		  if( (1 << mi) == C_MOUNTF ){
			if( bp != list ) *bp++ = ',';
			sym = id2sym0(tab,C_MOUNT);
			strcpy(bp,sym);
			bp += strlen(bp);
		  }
		}else
		if( flag & mask ){
			if( bp != list ) *bp++ = ',';
			sym = id2sym0(tab,flag);
			strcpy(bp,sym);
			bp += strlen(bp);
		}
	}
	*bp = 0;
}

static uconv0(convtype,tags,attr,negate)
{	int ri;
	int found;
	Attr *ap;
	char list[128];

	found = 0;
	for( ri = 0; ri < rewriteAttrX; ri++ ){
		ap = &rewriteAttr[ri];
		if( ap->a_context & tags )
		if( ap->a_id & attr ){
			found++;
			if( negate )
				ap->a_convtype &= ~convtype;
			else	ap->a_convtype |=  convtype;
		}
	}
	if( !negate && !found ){
		ap = &rewriteAttr[rewriteAttrX++];
		ap->a_id = attr;
		ap->a_context = tags;
		ap->a_convtype = convtype;

		mask2list(convsym,convtype,list);

		if( HTMLCONV_DEBUG )
		syslog_DEBUG("URICONV[%d] %s:%s/%s (%x/%x)\n",
			rewriteAttrX,
			list,
			id2sym(attrsym,attr),
			id2sym(elemsym,tags),
			attr,tags);
	}
}
static uconvx(convtype,tags,attr,negate)
{	int ti,ai;
	unsigned int amask,tmask;

	tmask = tags;
	for( ti = 0; tmask; ti++ ){
		if( tmask & 1 ){
			amask = attr;
			for( ai = 0; amask; ai++ ){
				if( amask & 1 )
					uconv0(convtype,(1<<ti),(1<<ai),negate);
				amask = (amask >> 1);
			}
		}
		tmask = (tmask >> 1);
	}
}
static uconvm(convtype,tags,attr,negate)
{	int lattr,ltags;
	Attr *ap;

	if( tags == E_ANY || attr == A_ANY ){
		uconv0(convtype,tags,attr,negate);
		return;
	}
	if( tags == E_DEFINEDS && attr == A_DEFINEDS ){
		for( ap = rewriteAttrDefault; ap->a_id; ap++ )
			if( ap->a_context == E_ANY )
				uconv0(convtype,ap->a_context,ap->a_id,negate);
			else	uconvx(convtype,ap->a_context,ap->a_id,negate);
		return;
	}
	if( tags == E_DEFINEDS && attr != A_DEFINEDS ){
		ltags = 0;
		for( ap = rewriteAttrDefault; ap->a_id; ap++ )
			if( ap->a_id == attr )
				ltags |= ap->a_context;
		tags = ltags;
	}else
	if( tags != E_DEFINEDS && attr == A_DEFINEDS ){
	}
	uconvx(convtype,tags,attr,negate);
}
static uconv1(convspec,convtype)
	char *convspec;
{	char sconv[256],sattr[256],stags[256];
	int attr,tags;
	int negate;

	if( *convspec == '!' || *convspec == '-' ){
		negate = 1;
		convspec++;
	}else	negate = 0;
	majorminorScan(convspec,sattr,stags);
	if( *sattr == 0 ) strcpy(sattr,"+");
	if( *stags == 0 ) strcpy(stags,"+");

	tags = sym2id(elemsym,stags);
	attr = sym2id(attrsym,sattr);

	uconvm(convtype,tags,attr,negate);
	return 0;
}

static tagmask1(stag,maskp)
	char *stag;
	int *maskp;
{	int itag;

	itag = sym2id(elemsym,stag);
	*maskp |= itag;
	return 0;
}
HTML_TagMask(stags)
	char *stags;
{	int itags;

	itags = 0;
	scan_commaListL(stags,0,tagmask1,&itags);
	return itags;
}
setKillTags(stags)
	char *stags;
{	int itags;

	itags = HTML_TagMask(stags);
	count_convs(C_KILL);
	uconvm(C_KILL,itags,A_END,0);
}

static setURIconvDefault(convtypes)
{	int ri;
	Attr *ap;

	if( HTMLCONV_DEBUG )
		syslog_DEBUG("HTMLCONV: setURIconvDefault\n");

	for( ri = rewriteAttrX; ri < NCONVS; ri++ ){
		ap = &rewriteAttrDefault[ri];
		if( ap->a_convtype == 0 )
			break;
		count_convs(ap->a_convtype);
		uconvm(ap->a_convtype,ap->a_context,ap->a_id,0);
		if( rewriteAttr[ri].a_id == 0 )
			break;
	}
}
setURICONVdefault(force){
	if( force || rewriteAttrInit == 0 ){
		setURIconvDefault(0);
		rewriteAttrInit = 1;
	}
}
static count_convs(convtype)
{
	if( convtype & C_MOUNTF	) URICONV_nMOUNTF++;
	if( convtype & C_MOUNTR	) URICONV_nMOUNTR++;
	if( convtype & C_MOUNTD	) URICONV_nMOUNTD++;
	if( convtype & C_MOUNT	) URICONV_nMOUNT++;
	if( convtype & C_PARTIAL) URICONV_nPARTIAL++;
	if( convtype & C_FULL	) URICONV_nFULL++;
	if( convtype & C_KILL   ) TAGCONV_nKILL++;
	if( convtype & C_NORMAL	) URICONV_nNORMAL++;
}
setURIconv(conv)
	char *conv;
{	char sconvtype[128],convlist[1024];
	int attr;
	int convtype;

	rewriteAttrInit++;
	fieldScan(conv,sconvtype,convlist);
	if( *conv == 0 ){
		/* URICONV="" means clearing URICONV */
		return;
	}
	if( *sconvtype == 0 ) strcpy(sconvtype,"+");
	if( *convlist  == 0 ) strcpy(convlist,"+");

	convtype = list2mask(convsym,sconvtype);
	count_convs(convtype);
	scan_commaListL(convlist,0,uconv1,convtype);
}

static sym2tid(tag)
	char *tag;
{	char tsym[32],*op;
	unsigned char *tp;
	char *ox;
	int tid;

/*
	tid = E_ANY;
*/
	tid = E_UNKNOWN;
	if( tag != 0 ){
		tp = tag;
		if( *tp == '<' ){
			tp++;
			if( *tp == '/' )
			{
/* | E_END */
				tp++;
			}
		}
		while( isspace(*tp) )
			tp++;
		op = tsym;
		ox = tsym + sizeof(tsym) - 1;
		while( isAlpha(*tp) )
		{	if( ox <= op )
				break;
			*op++ = *tp++;
		}
		*op = 0;
		tid = sym2id(elemsym,tsym);
		if( tid == 0 )
/*
			tid = E_ANY;
*/
			tid = E_UNKNOWN;
	}
	return tid;
}

HTML_attrTobeConv(attr,tag,uconvp)
	char *attr,*tag;
	int *uconvp;
{	int ai,ri;
	char asym[32];
	int tid,aid,alen,aidmask;
	int uconv,ruconv;
	Attr *rp;
	int altno;

	setURICONVdefault(0);
	alen = scanNameScreen(attrsym,&attrNames,attr,asym);
	if( alen <= 0 )
		return 0;
	aid = sym2id(attrsym,asym);

	/*
	 *	Searching a rule for the attribute
	 */
	if( uconvp ){
		uconv = *uconvp;
		*uconvp = 0;
	}else	uconv = URICONV_ANY;

	attrSerno++;
	altno = 0;
	tid = 0;

	for( ri = 0; ri < rewriteAttrX; ri++ ){
		rp = &rewriteAttr[ri];
		aidmask = rp->a_id;
		/* aid is 0 if not in A_DEFINEDS */
		if( (aidmask & aid) || aidmask == A_ANY ){
			if( tid == 0 )
				tid = sym2tid(tag);
			if( HTMLCONV_DEBUG ){
				syslog_DEBUG(
				"URICONV:%4d.%d %s <%-6s %-6s>[%5x/%5x]\n",
				attrSerno,++altno,
				rp->a_context & tid ? "*":" ",
				id2sym(elemsym,tid),
				id2sym(attrsym,rp->a_id),
				rp->a_context,tid);
			}
			if( rp->a_context & tid ){
				ruconv = rp->a_convtype;
				if( ruconv & uconv ){
					if( uconvp )
						*uconvp = ruconv;
					return alen;
				}
			}
		}
	}
	return 0;
}

dumpHTMLCONV(){
	char buff[4096],*bp,*sym;
	int si,ri,ci,aid,ctx,nput;
	Attr *rp;

	bp = buff;
	if( ENCODE_HTML_ENTITIES ){
		strcpy(bp,"HTMLCONV=enent\n");
		bp += strlen(bp);
	}
	if( DECODE_HTML_ENTITIES ){
		strcpy(bp,"HTMLCONV=deent\n");
		bp += strlen(bp);
	}

	setURICONVdefault(0);

	bp += strlen(strcpy(bp,"URICONV=defelem:{"));
	for( si = 2; sym = elemsym[si]._sym; si++ ){
		if( 2 < si ) *bp++ = ',';
		bp += strlen(strcpy(bp,sym));
	}
	*bp++ = '}';
	*bp++ = '\n';
	bp += strlen(strcpy(bp,"URICONV=defattr:{"));
	for( si = 2; sym = attrsym[si]._sym; si++ ){
		if( 2 < si ) *bp++ = ',';
		bp += strlen(strcpy(bp,sym));
	}
	*bp++ = '}';
	*bp++ = '\n';
	for( ri = 0; ri < rewriteAttrX; ri++ ){
		rp = &rewriteAttr[ri];
		if( rp->a_convtype == 0 )
			continue;
		sprintf(bp,"URICONV=");
		bp += strlen(bp);
		mask2list(convsym,rp->a_convtype,bp);
		bp += strlen(bp);
		*bp++ = ':';

		aid = rp->a_id;
		ctx = rp->a_context;
		if( ctx == E_ANY ){
			sprintf(bp,"%s/*",id2sym(attrsym,aid));
			bp += strlen(bp);
		}else{
		  nput = 0;
		  for( ci = 0; ci < 32; ci++ ){
		    if( (1 << ci) & ctx ){
			if( nput != 0 ) *bp++ = ',';
			sprintf(bp,"%s/%s",
			id2sym(attrsym,aid),id2sym(elemsym,(1<<ci)));
			bp += strlen(bp);
			nput++;
		    }
		  }
		}
		*bp++ = '\n';
	}
	*bp = 0;
	syslog_ERROR("\n### HTMLCONV configuration:\n%s",buff);
	return;
}

/*######################################################################*/
static struct {
	char	*ee;
	char	 de;
} entities[] = {
	{"&lt;",	'<'},
	{"&gt;",	'>'},
	{"&amp;",	'&'},
	{"&quot;",	'"'},
	0
};

isHTMLentity(str,chp)
	char *str;
	int *chp;
{	int ei,ej;
	char *es,ec,*sp;
	int off;

	off = *str == '&' ? 0 : 1;
	for( ei = 0; es = entities[ei].ee; ei++ ){
		for( ej = 0; ec = es[off+ej]; ej++ ){
			if( ec != str[ej] )
				break;
			if( ec == ';' ){
				*chp = entities[ei].de;
				return ej+1;
			}
		}
	}
	return 0;
}

/*
 *  Decode mal-encoded entities in ISO-2022 multibyte charset text.
 */
decode_entities(src,dst)
	char *src,*dst;
{
	if( DECODE_HTML_ENTITIES == 0 ){
		strcpy(dst,src);
		return;
	}
	decodeEntities(src,dst,0);
}
decodeEntities(src,dst,anywhere)
	char *src,*dst;
{	char *sp,*dp;
	int ei,len;
	char *ee;
	int in2byte;

	in2byte = 0;
	dp = dst;
	for( sp = src; *sp; )
	{
		if( *sp == 033 ){
			switch( sp[1] ){
				case '$': in2byte  = 1; break;
				case '(': in2byte  = 0; break;
			}
		}else
		if( (anywhere||in2byte) && *sp == '&' ){
			for( ei = 0; ee = entities[ei].ee; ei++ ){
				len = strlen(ee);
				if( strncmp(sp,ee,len) == 0 ){
					*dp++ = entities[ei].de;
					sp += len;
					goto Next;
				}
			}
		}
		*dp++ = *sp++;
	Next:;
	}
	*dp = 0;
}

encode_entitiesX(src,dst,siz)
	char *src,*dst;
{
	if( ENCODE_HTML_ENTITIES == 0 ){
		Strncpy(dst,src,siz);
		return;
	}
	encodeEntitiesX(src,dst,siz);
}
encodeEntitiesX(src,dst,siz)
	char *src,*dst;
{	int in2byte;
	char *sp,*dp;
	char de;
	int ei,len;
	char *xp;
	int nenc;

	nenc = 0;
	in2byte = 0;
	dp = dst;
	xp = dst + (siz - 1);
	for( sp = src; *sp; sp++ ){
		if( xp <= dp )
			break;
		if( *sp == 033 ){
			switch( sp[1] ){
				case '$': in2byte  = 1; break;
				case '(': in2byte  = 0; break;
			}
		}else
		if( !in2byte ){
			for( ei = 0; de = entities[ei].de; ei++ ){
				if( *sp == de ){
					strcpy(dp,entities[ei].ee);
					dp += strlen(dp);
					nenc++;
					goto Next;
				}
			}
		}
		*dp++ = *sp;
	Next:;
	}
	*dp = 0;
	return nenc;
}

char *strpbrk1B(src,brks,dst)
	char *src,*brks,**dst;
{	char *sp,sch,*dp,swch,bch;
	int si,bi;
	int in2byte = 0;
	int found = 0;

	dp = *dst;
	for( sp = src; sch = *sp; ){
		if( sch == 033 ){
			swch = sp[1];
			if( (swch == '$' || swch == '(') && sp[2] != 0 ){
				for( si = 0; si < 3; si++ )
					*dp++ = *sp++;
				if( swch == '$' ) in2byte = 1; else
				if( swch == '(' ) in2byte = 0;
				continue;
			}
		}
		if( !in2byte ){
			for( bi = 0; bch = brks[bi]; bi++ )
				if( sch == bch ){
					found = 1;
					goto FOUND;
				}
		}
		*dp++ = sch;
		sp++;
	}
FOUND:
	*dp = 0;
	*dst = dp;

	if( found )
		return sp;
	else	return 0;
}

scan_html(src,dst,stack)
	char *src,*dst,*stack;
{
}
