/*////////////////////////////////////////////////////////////////////////
Copyright (c) 1994-1999 Yutaka Sato
Copyright (c) 1994-1999 Electrotechnical Laboratry (ETL), AIST, MITI

Permission to use, copy, and distribute this material for any purpose
and without fee is hereby granted, provided that the above copyright
notice and this permission notice appear in all copies.
ETL MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR SUITABILITY OF THIS
MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS", WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES.
/////////////////////////////////////////////////////////////////////////
Content-Type:	program/C; charset=US-ASCII
Program:	hostlist.c
Author:		Yutaka Sato <ysato@etl.go.jp>
Description:

    Each element in HostList is one of follows:

	IP-address matching with mask:
		host/255.255.255.0
		host/@C

	IP-address range:
		192.31.200.[65-94]
		192.31.[197-206]
		150.29.[0-255]

	Domain name pattern macthing by name:
		*.ac.jp
		*.jp

    A special host-name "?" matches with hosts which name cannot be
    retrieved with DNS, thus unknown.

    Each element can be prefixed with UserList.
    A special user-name "?" matches when the user cannot be
    identified with IDENTd.

    Special elements which add a flavor of composit operation :-)

	AND: A,&,B -- if A is false, false as a whole (can be "&{A,B}" ???)
	OR : A,|,B -- if A is true, ture as a whole (short cut evaluation)
	NEG: A,!,B -- negate the result of A, then evaluate B (can be "!{A,B}")

History:
	940623	added simple masking (suggested by <yoko@huie.hokudai.ac.jp>)
	940807	separated from access.c
//////////////////////////////////////////////////////////////////////#*/
#include "ystring.h"
#include <ctype.h>

#include "hostlist.h"
#define FILE void

#include "log.h"
#undef Verbose
#define Verbose (LOG_type&L_HOSTMATCH)==0?0:putLog0

extern char *frex_create(),*frex_match();
extern char *VA_inAddr();
#define inAddr(addr)	VA_inAddr(addr)
char *hostmatch_asisaddr = "/a";
char *hostmatch_withauth = "/u";
char *hostmatch_ignauth = "*";

#define HL_INCSIZE	8
#define HL_MAXDEPTH	10

static char opsyms[] = "+!-?";
#define ON_EQ	0	/* ON  if matched */
#define ON_NEQ	1	/* ON  if not matched */
#define OFF_EQ	2	/* OFF if matched */
#define OFF_NEQ	3	/* OFF if not matched */

#define OP_AND	4
#define OP_OR	5
#define OP_NEG	6
#define OP_HOUR	7

#define	DO_NOP	0
#define	DO_ON	1
#define DO_OFF	2

typedef struct {
	double	currentTime;
	VAddr	clientsideHost;
	VAddr	clientSockHost;
	VAddr	clientHost;
	char	agentname[128];
} ClientInfo;
typedef struct {
	int	ht_size;
	int	ht_id;
	int	ht_hash;
     HostList **ht_list;
} HostListTab;
typedef struct {
	HostList	he_lexHosts;
	HostListTab	he_listTab;
	ClientInfo	he_clientInfo;
} HostListEnv;
static HostListEnv *hostListEnv;
static char LexHosts[] = "LexHosts";
#define lexHosts	(&hostListEnv->he_lexHosts)
#define hostListTab	hostListEnv->he_listTab
#define hostListTabHash	hostListEnv->he_listTab.ht_hash
#define hostListName(id) hostListEnv->he_listTab.ht_list[id]->hl_what
minit_hostlist()
{
	if( hostListEnv == 0 )
	{
		hostListEnv = NewStruct(HostListEnv);
		hostListEnv->he_lexHosts.hl_what = LexHosts;
	}
}
extern addHostList1();
extern char *DELEGATE_LOCALNET;
static resize_hostlisttab(ht)
	HostListTab *ht;
{	int newsize,li;
	HostListTab oht;
	HostList *HL;

	oht = *ht;
	if( oht.ht_size ){
		newsize = oht.ht_size * 2;
	}else	newsize = 16;
	Verbose("HOSTLIST resized [%d -> %d]\n",oht.ht_size,newsize);
	ht->ht_size = newsize;
	ht->ht_list = (HostList**)calloc(sizeof(HostList*),newsize);
	ht->ht_hash = strid_create(newsize*2);
	if( oht.ht_size ){
		for( li = 1; li < oht.ht_size; li++ ){
			ht->ht_list[li] = HL = oht.ht_list[li];
			strid(ht->ht_hash,HL->hl_what,li);
		}
		/* free(oht.ht_list); might be referred by self redefinition */
	}
}
putHostListTab(aname,HL)
	char *aname;
	HostList *HL;
{	int listid;

	listid = strid(hostListTabHash,aname,-1);
	if( listid < 0 ){
		listid = ++hostListTab.ht_id;
		strid(hostListTabHash,aname,listid);
		hostListTab.ht_list[listid] = HL;
	}
}
addHostListTab(dname,dhostlist)
	char *dname,*dhostlist;
{	int listid,apptype,applist,noresolv;
	char *aname,anameb[64],*aflags,*ahostlist;
	HostList *HL,OHL;

	if( noresolv = *dname == '-' )
		dname++;
	if( aflags = strchr(dname,'/') ){
		wordscanY(dname,anameb,sizeof(anameb),"^/");
		aname = anameb;
		aflags++;
	}else{
		aname = dname;
		aflags = "";
	}
	apptype = strchr(aflags,'+') != NULL;
	applist = strncmp(dhostlist,"+,",2) == 0;
	if( applist )
		ahostlist = dhostlist + 2;
	else	ahostlist = dhostlist;

	listid = strid(hostListTabHash,aname,-1);
	if( 0 < listid ){
		HL = hostListTab.ht_list[listid];
		if( HL->hl_flags & HL_PROTECT ){
			sv1log("ERROR HOSTLIST=%s protected\n",aname);
		}else
		if( (HL->hl_flags & HL_APPEND) || apptype || applist ){
		}else{
			OHL = *HL;
			bzero(HL,sizeof(HostList));
			HL->hl_what = OHL.hl_what;
			HL->hl_flags = OHL.hl_flags;
			sv1log("[%d] HOSLIST=%s overwrite\n",listid,aname);
		}
	}else{
		listid = ++hostListTab.ht_id;
		if( hostListTab.ht_size <= listid )
			resize_hostlisttab(&hostListTab);
		strid(hostListTabHash,aname,listid);
		HL = NewStruct(HostList);
		HL->hl_what = StrAlloc(aname);
		if( apptype )
			HL->hl_flags |= HL_APPEND;
		hostListTab.ht_list[listid] = HL;
	}
	if( strchr(aflags,'p') )
		HL->hl_flags |= HL_PROTECT;
	if( strchr(aflags,'c') )
		HL->hl_flags |= HL_BYCLHOST;
	if( strchr(aflags,'i') )
		HL->hl_flags |= HL_NOIDENT;
	if( strchr(aflags,'r') || noresolv )
		HL->hl_flags |= HL_NORESOLV;
	if( strchr(aflags,'a') )
		HL->hl_flags |= HL_BYADDR;
	if( strchr(aflags,'u') )
		HL->hl_flags |= HL_BYAUTH;
	if( strchr(aflags,'A') )
		HL->hl_flags |= HL_BYAGENT;
	scan_commaListL(ahostlist,0,addHostList1,HL);
	Verbose("[%d] HOSLIST=%s:%s\n",listid,dname,dhostlist);
}
static init_hostListTab()
{
	if( hostListTabHash == 0 ){
		resize_hostlisttab(&hostListTab);
		addHostListTab(".localnet",DELEGATE_LOCALNET);
		addHostListTab(".RELIABLE","*");
		addHostListTab(".socksdst","!.localnet");
	}
}
scan_HOSTLIST(ctx,listdef)
	void *ctx;
	char *listdef;
{	char *dp,name[32];

	dp = strchr(listdef,':');
	if( dp == 0 ){
		return;
	}
	wordscanY(listdef,name,sizeof(name),"^:");
	init_hostListTab();
	addHostListTab(name,dp+1);
}

#define clientInfo	hostListEnv->he_clientInfo
#define CurrentTime	clientInfo.currentTime
#define ClientsideHost	clientInfo.clientsideHost
#define ClientSockHost	clientInfo.clientSockHost
#define ClientHost	clientInfo.clientHost
#define ClientAgent	clientInfo.agentname
VA_HL_pushClientInfo(Now,peerhost,sockhost)
	double Now;
	VAddr *peerhost,*sockhost;
{
	CurrentTime = Now;
	ClientsideHost = *peerhost;
	ClientSockHost = *sockhost;
}
HL_setClientAgent(agent)
	char *agent;
{
	lineScan(agent,clientInfo.agentname);
}
HL_setClientIF(addr,port,remote)
	char *addr;
{
	if( addr )
		VA_setVAddr(&ClientSockHost,addr,port,remote);
	else	ClientSockHost = AddrNull;
}
HL_popClientInfo()
{
	ClientsideHost = AddrNull;
	ClientSockHost = AddrNull;
}
HL_setClientInfo(peerhost)
	VAddr *peerhost;
{
	ClientHost = *peerhost;
}

static matchProto(protolist,proto)
	char *protolist[],*proto;
{	int isin,pi;
	char *proto1;
	int op,match;

	if( *protolist[0] == '!' )
		isin = 1;
	else	isin = 0;

	for( pi = 0; proto1 = protolist[pi]; pi++ ){
		if( *proto1 == '!' ){
			proto1++;
			op = OFF_EQ;
		}else	op = ON_EQ;

		if( streq(proto1,"*") )
			match = 1;
		else	match = streq(proto1,proto);

		if( match && op == ON_EQ )	isin = 1; else
		if( match && op == OFF_EQ )	isin = 0;
		Verbose("--> %d : %d : %s : %s : %s\n",
			isin, match, op==OFF_EQ?"NEQ":"EQ",
			proto, proto1);
	}
	return isin;
}
static char **makeUserRexp(uc,userlist)
	char *userlist[];
{	char **rexpv;
	char *user;
	int ui;

	rexpv = (char**)StructAlloc(sizeof(char*)*uc);
	for( ui = 0; ui < uc; ui++ ){
		user = userlist[ui];
		if( *user == '!' )
			user++;
		if( !streq(user,"*") && strpbrk(user,"*[") ){
			sv1log("##REGEXP COMPILED: %s\n",user);
			rexpv[ui] = frex_create(user);
		}else	rexpv[ui] = 0;
	}
	return rexpv;
}
static matchUser(userlist,rexplist,username)
	char *userlist[];
	char *rexplist[];
	char *username;
{	int isin,ui;
	char *user;
	char *rexp,*tail;
	int op,match;

	if( *userlist[0] == '!' )
		isin = 1;
	else	isin = 0;

	for( ui = 0; user = userlist[ui]; ui++ ){
		if( *user == '!' ){
			user++;
			op = OFF_EQ;
		}else	op = ON_EQ;

		if( streq(user,"*") )
			match = 1;
		else
		if( rexp = rexplist[ui] ){
			tail = frex_match(rexp,username);
			match = tail && *tail == 0;
			sv1log("##REGEXP EXECUTED ==> %d : %s : %s\n",
				match,username,user);
		}
		else	match = strcmp(user,username) == 0;

		if( match && op == ON_EQ )	isin = 1; else
		if( match && op == OFF_EQ )	isin = 0;
		Verbose("--> %d : %d : %s : %s : %s\n",
			isin, match, op==OFF_EQ?"NEQ":"EQ",
			username,user);
	}
	return isin;
}
static matchPort(portlist,portnum)
	char *portlist[];
{	int isin,pi;
	char *ports;
	int op,match;
	int port0,port1;

	if( *portlist[0] == '!' )
		isin = 1;
	else	isin = 0;

	for( pi = 0; ports = portlist[pi]; pi++ ){
		if( *ports == '!' ){
			ports++;
			op = OFF_EQ;
		}else	op = ON_EQ;

		if( streq(ports,"*") )
			match = 1;
		else{
			switch( sscanf(ports,"%d-%d",&port0,&port1) ){
			case 0: match = 0; break;
			case 1: match = port0 == portnum; break;
			case 2: match = port0 <= portnum && portnum <= port1;
				break;
			}
		}

		if( match && op == ON_EQ )	isin = 1; else
		if( match && op == OFF_EQ )	isin = 0;
		Verbose("--> %d : %d : %s : %d : %s\n",
			isin, match, op==OFF_EQ?"NEQ":"EQ",
			portnum,ports);
	}
	return isin;
}

static match_addrrange(Hp,Vaddr)
	Host *Hp;
	VAddr *Vaddr;
{	int ai;
	int addr = Vaddr->I3;
	Uchar a1;

	for(ai = 0; ai < 4; ai++){
		a1 = (addr >> ((3-ai)*8)) & 0xFF;
		if( a1 < Hp->ranges[ai].h_low || Hp->ranges[ai].h_high < a1 )
			return 0;
	}
	return 1;
}

/* RFC1597 private IP address */
#define PRIVATEC	((192<<24)|(168<<16))
#define PRIVATEBL	((172<<24)|(16<<16))
#define PRIVATEBH	((172<<24)|(31<<16))
#define PRIVATEA	(10<<24)

static default_netmask(Mask,Vaddr)
	VAddr *Mask,*Vaddr;
{	unsigned int ip4c,mask1;
	unsigned int ip4,net;

	if( Mask->I3 != CLASS_MASK )
	if( Mask->I3 != HEURISTIC_MASK )
		return 0;

	ip4c = (Vaddr->I3 >> 24) & 0xFF;
	if( ip4c < 128 ) mask1 = 0xFF000000; else
	if( ip4c < 192 ) mask1 = 0xFFFF0000; else
	       mask1 = 0xFFFFFF00;

	if( Mask->I3 == HEURISTIC_MASK ){
		ip4 = Vaddr->I3;
		net = ip4 & mask1;
		if( net == PRIVATEC 
		 || PRIVATEBL <= net && net <= PRIVATEBH
		 || net == PRIVATEA ){
			/* use class mask for private address */ 
		}else
		if( ip4c < 128 )/* class A */
		{
			mask1 = 0xFFFFFF00;
			*Mask = MaskZero;
			Mask->I3 = mask1;
			sv1log("default netmask %s/. = %X\n",
				VA_inAddr(Vaddr),mask1);
		}
	}

	*Mask = MaskZero;
	Mask->I3 = mask1;
	return 1;
}

static WHMtoi(st,dflt)
	char *st,*dflt;
{	char buf[8],sc;
	int si;

	strcpy(buf,dflt);
	for( si = 0; sc = st[si]; si++ )
		buf[si] = sc;
	return atoi(buf);
}
static matchTimePeriod(period,mask)
	char *period;
{	char hour[8];
	int hc,h1,h2;
	int mc,m1,m2;
	char *pp,*fmt,*smin,*smax,sh1[32],sh2[32];
	int min,max,mod;

	/*
	 * -T.000-499/1000
	 * -T.0-4/10
	 */
	if( 0 < mask ){
		sscanf(period,"%d-%d",&m1,&m2);
		mc = (int)((CurrentTime - (int)CurrentTime) * mask) % mask;
		if( m1 <= mc && mc <= m2 )
			return 1;
		else	return 0;
	}

	pp = period;
	switch( *pp ){
		case 'w': fmt = "%w%H%M"; smin = "00000"; smax = "62359"; pp++;
			  mod = 70000; break;
		case 'u': fmt = "%u%H%M"; smin = "10000"; smax = "72359"; pp++;
			  mod = 80000; break;
		default:  fmt =   "%H%M"; smin =  "0000"; smax =  "2359";
			  mod = 10000; break;
	}
	switch( sscanf(pp,"%[0-9]-%[0-9]",sh1,sh2) ){
		case 0: return 0;
		case 1: h1 = WHMtoi(sh1,smin); h2 = WHMtoi(sh1,smax); break;
		case 2: h1 = WHMtoi(sh1,smin); h2 = WHMtoi(sh2,smax); break;
	}
	min = atoi(smin);
	max = atoi(smax);
	h1 = h1 % mod;
	h2 = h2 % mod;

	StrftimeLocal(hour,sizeof(hour),fmt,(int)CurrentTime);
	hc = atoi(hour);

	Verbose("-T %d < %d-[%d]-%d < %d ?\n",min,h1,hc,h2,max);

	if( h1 <= h2 ){
		if( h1 <= hc && hc <= h2 )
			return 1; 
	}else{
		if( h1 <= hc && hc <= max || min <= hc && hc <= h2 )
			return 1;
	}
	return 0;
}

static notResolved(Vaddr,name)
	VAddr *Vaddr;
	char *name;
{
	if( AddrInvalid(*Vaddr) )
		return 1;

	if( isinetAddr(name) != 0 ) /* name is in "dd.dd.dd.dd" IP-address */
		return 1;
	return 0;
}
static char *scanVdomain(hostname,hostnameb,vdomain)
	char *hostname,*hostnameb,*vdomain;
{	char *vdom,*vp,*sp,*dp,ch;

	*vdomain = 0;
	vdom = strstr(hostname,".-");
	if( vdom == 0 )
		return 0;

	strcpy(hostnameb,hostname);
	hostname = hostnameb;
	vp = sp = strstr(hostname,".-");
	dp = vdomain;
	for( vp += 2; ch = *vp; vp++ ){
		if( isalnum(ch) )
			*dp++ = ch;
	}
	if( dp == vdomain )
		return 0;

	*dp = 0;
	while( *sp++ = *vp++ );
	return vdom;
}

#define HostOutgoingIF(hp)	(hp[0]=='.' && hp[1]=='o' && hp[2]==0)
#define HostIncomingIF(hp)	(hp[0]=='.' && hp[1]=='i' && hp[2]==0\
			      || hp[0]=='-' && hp[1]==0)
#define HostIncomingIFX(hp)	(hp[0]=='-' && hp[1]=='P')
#define HostOnClientside(hp)	(hp[0]=='-' && hp[1]=='C' && hp[2]==0\
			      || hp[0]=='.' && hp[1]=='C' && hp[2]==0)
#define TimePeriod(hp)		(hp[0]=='-' && hp[1]=='T' && hp[2]=='.')
#define HostUnknown(hp)		(hp[0]=='?' && hp[1]==0)
#define HostMatchExactly(hp,hn)	(strcasecmp(hp,hn) == 0)

static VA_hostIsinListX(hostlist,proto,cname,alias,Vaddr,username,ac,av,lev)
	HostList *hostlist;
	char *proto,*cname,*alias;
	VAddr *Vaddr;
	AuthInfo *av[];
	char *username;
{	int portnum = Vaddr->a_port;
	int hi,nhosts;
	VAddr hosti,mask,ahost,chost;
	Host *Hp;
	char *hostpat;
	int op;
	int isin,match,do_onoff;
	char hostnameb[256],*vdom,vdomb[128],*vdom1;
	char *hostname;
	char *domhost;
	char *scname = cname;
	char *salias = alias;
	int byauth;
	int byagent;

	hosti = *Vaddr;

	if( hostlist->hl_flags & HL_BYCLHOST ){
		hosti = ClientHost;
		cname = ClientHost.a_name;
		alias = 0;
		Verbose("BYCLHOST: %s %s\n",hostlist->hl_what,cname);
	}
	if( byauth = (hostlist->hl_flags & HL_BYAUTH) ){
		if( ac == 0 ){
			Verbose("BYAUTH[%s]: no authinfo: %s\n",
				hostlist->hl_what,cname);
			return 0;
		}
	}
	byagent = hostlist->hl_flags & HL_BYAGENT;

	if( scanVdomain(cname,hostnameb,vdomb) ){
		cname = hostnameb;
		vdom = vdomb;
	}else	vdom = 0;

	strtolower(cname,hostnameb);
	cname = hostnameb;

	op = 0;
	isin = 0;
	nhosts = hostlist->hl_cnt;

	for( hi = 1; hi <= nhosts; hi++ ){
		Hp = hostlist->hl_List[hi];
		op = Hp->h_op;
		if( op == OP_AND ){
			if( isin ){
				Verbose("[%d/%d] &=> 1 AND succeed, cont.\n",
					hi,nhosts);
				isin = 0;
				continue;
			}else{
				Verbose("[%d/%d] &=> 0 AND failed, break\n",
					hi,nhosts);
				for( hi++; hi <= nhosts; hi++ ){
					Hp = hostlist->hl_List[hi];
					op = Hp->h_op;
					if( Hp->h_op == OP_OR )
						break;
				}
				if( op != OP_OR )
					break;
				Verbose("[%d/%d] |=> found OR.\n",hi,nhosts);
				continue;
			}
		}
		if( op == OP_OR ){
			if( isin ){
				Verbose("[%d] |=> OR 1 succeed, break\n",hi);
				break;
			}else{
				Verbose("[%d] |=> OR 0 ineffective, cont.\n",hi);
				continue;
			}
		}
		if( op == OP_NEG ){
			isin = !isin;
			Verbose("[%d/%d] !=> %d NEGATE\n",hi,nhosts,isin);
			continue;
		}

		vdom1 = Hp->h_vdomain;
		if( vdom == 0 && vdom1 != 0
		 || vdom != 0 && vdom1 == 0
		 || vdom != 0 && vdom1 != 0 && !streq(vdom,vdom1) )
			continue;

		if( isin && (op == ON_EQ || op == ON_NEQ) ||
		   !isin && (op ==OFF_EQ || op==OFF_NEQ) ){
			/* nothing will happen */
			continue;
		}
		match = 0;

		if( byagent || Hp->h_type == HT_BYAGENT ){
			match = strstr(ClientAgent,Hp->h_name) != 0;
			Verbose("User-Agent: %s ==> %d\n",Hp->h_name,match);
			goto ACCUM;
		}

		mask = Hp->h_mask;
		hostpat = Hp->h_name;

		domhost = 0;
		if( hostpat[0] == '*' && hostpat[1] == '.' )
			domhost = hostpat + 2;

		if( Hp->h_asis && alias )
			hostname = alias;
		else	hostname = cname;

		if( Hp->h_listid ){
			int listid = Hp->h_listid;
			if( HL_MAXDEPTH < lev ){
				sv1log("ERROR: HostList too deep (%d): %s\n",
					lev,hostlist->hl_what);
				return 0;
			}
			match = VA_hostIsinListX(hostListTab.ht_list[listid],
				proto,scname,salias,Vaddr,username,0,NULL,lev+1);
			Verbose("[%d/%d] HOSTLIST[%d]:%s(%s) -> %d\n",hi,nhosts,
				listid,Hp->h_name,hostListName(listid),match);
			goto USERPORT;
		}
		if( TimePeriod(hostpat) ){
			match = matchTimePeriod(hostpat+3,mask.I3);
			Verbose("[%d/%d] TimePeriod = %s / %d = %d\n",
				hi,nhosts,hostpat,mask.I3,match);
			goto ACCUM;
		}

		if( HostOnClientside(hostpat) ){
			Verbose("[%d/%d] -C (ClientHost) = %s\n",hi,nhosts,
				inAddr(&ClientsideHost));
			default_netmask(&mask,&ClientsideHost);
			AddrAND(chost,mask,ClientsideHost);
		}else
		if( HostOutgoingIF(hostpat) ){
			VAddr ohost;
			if( AddrInvalid(hosti) ){
				chost = hosti; /* any different from hosti */
				chost.I3 = ~chost.I3;
				ohost = chost;
				mask = AddrNull;
			}else{
				default_netmask(&mask,&hosti);
				VA_hostIFto(&hosti,&mask,&ohost);
				AddrAND(chost,ohost,mask);
			}
			Verbose("[%d/%d] .o (OutgoingIF) = %s -> %s\n",hi,nhosts,
				inAddr(&ohost),inAddr(&hosti));
		}else
		if( Hp->h_type & HT_CLIF ){
			Verbose("[%d/%d] -P (IncomingIFX) = %s\n",hi,nhosts,
				inAddr(&ClientSockHost));
			default_netmask(&mask,&ClientSockHost);
			AddrAND(chost,mask,ClientSockHost);
			/*
			 * -P<port> alwayes checks if the incomming port is
			 * -P<port>, unconditionally on what host is tested
			 */
			hosti = ClientSockHost;
			hostname = ClientSockHost.a_name;
			portnum = ClientSockHost.a_port;
		}else
		if( HostIncomingIF(hostpat) ){
			Verbose("[%d/%d] .i (IncomingIF) = %s\n",hi,nhosts,
				inAddr(&ClientSockHost));
			default_netmask(&mask,&ClientSockHost);
			AddrAND(chost,mask,ClientSockHost);
		}else	AddrAND(chost,mask,Hp->h_addr);
		AddrAND(ahost,hosti,mask);

		/*
		if( byauth ){
		*/
		if( byauth && (Hp->h_type&HT_CLIF)==0 ){
			goto AUTHMATCH;
		}
		if( (Hp->h_type & HT_BYHOST) == 0 ){
			goto EXIT_BYHOST;
		}
		if( HostUnknown(hostpat) && notResolved(&hosti,hostname) ){
			Verbose("UNKNOWN HOST\n");
			match = 1;
		}else
		if( domhost && HostMatchExactly(domhost,hostname) ){
			Verbose("[%d/%d] HOST==DOMAIN MATCH: %s %c= %s ?\n",
				hi,nhosts,hostname,opsyms[op],hostpat);
			match = 1;
		}else
		if( HostMatchExactly(hostpat,hostname) ){
			Verbose("[%d/%d] EXACT NAME MATCH: %s %c= %s ?\n",
				hi,nhosts,hostname,opsyms[op],hostpat);
			match = 1;
		}else
		if( !AddrEQ(ahost,AddrZero) && !AddrEQ(chost,AddrZero) ){
			Verbose("[%d/%d] ADDR MATCH: %s %c= %s ?\n",hi,nhosts,
				inAddr(&ahost),opsyms[op],inAddr(&chost));
			match = AddrEQ(ahost,chost);
		}else
		if( !AddrEQ(ahost,AddrZero) && Hp->ranges[0].h_high ){
			Verbose("[%d/%d] ADDR RANGE: %s[%s] %c= %s ?\n",hi,nhosts,
				hostname,inAddr(&ahost),opsyms[op],hostpat);
			match = match_addrrange(Hp,&ahost);
		}else
		if( !match ){
			Verbose("[%d/%d] REGEXP NAME MATCHING: %s %c= %s ?\n",
				hi,nhosts,hostname,opsyms[op],hostpat);
			match = rexpmatch(hostpat,hostname)
		     		|| streq(hostpat,hostname);

			if( !match )
			if( hostname != alias )
			if( alias != NULL )
			if( !strcaseeq(alias,cname) ){
				Verbose("[%d/%d] ALIAS MATCHING: %s %c= %s ?\n",
					hi,nhosts,alias,opsyms[op],hostpat);
				match = rexpmatch(hostpat,alias)
					|| strcaseeq(hostpat,alias);

				if( !match )
				if( domhost )
					match = strcaseeq(domhost,alias);
			}
		}
USERPORT:
		if( match && Hp->h_proto ){
			if( proto != ANYP && !matchProto(Hp->h_proto,proto) ){
				Verbose("--> proto name unmatch: %s\n",proto);
				match = 0;
			}
		}
		if( match && Hp->h_port ){
			if( portnum && !matchPort(Hp->h_port,portnum) ){
				Verbose("--> port num. unmatch: %d\n",portnum);
				match = 0;
			}
		}
		if( match && Hp->h_user ){
			if( username == 0 )
				username = "?";
			if( !matchUser(Hp->h_user,Hp->h_userRexp,username) ){
				Verbose("--> user name unmatch: %s\n",username);
				match = 0;
			}
		}

EXIT_BYHOST:
		/*
		if( !match && Hp->h_user
		*/
		if( !match && HT_BYUSER(Hp) && ac <= 0
		 && username && streq(username,hostmatch_ignauth) ){
			/* postpone matching until auth. info. is got.
			 * "hostmatch_ignauth" might be replaced with "ac == 0"
			 */
			match = 1;
		}else
		/*
		 * + /u,hostList ... byauth
		 * + user@host
		 * + -a/authhost
		 * + -c/domain
		 *
		if( !match && (byauth || Hp->h_user) && 0 < ac ){
		*/
		if( !match && (byauth || HT_BYUSER(Hp)) && 0 < ac ){
			int ai;
			AuthInfo *ap;
		AUTHMATCH:
			for( ai = 0; ai < ac; ai++ ){
				int type;
				ap = av[ai];
				if( ap->i_stat & AUTH_SET )
					type = HT_BYAUTH;
				else	type = HT_BYCERT;
				if( (Hp->h_type & type) == 0 ){
					continue;
				}

				if( Hp->h_user )
				match = matchUser(Hp->h_user,Hp->h_userRexp,ap->i_user);
				else	match = 1;
				if( match ){
					match = streq(hostpat,ap->i_Host)
					     || rexpmatch(hostpat,ap->i_Host)
					     || hostcmp(hostpat,ap->i_Host) == 0 ;
				}
				if( match && Hp->h_port && 0 < ap->i_Port )
				if( !matchPort(Hp->h_port,ap->i_Port) ){
					Verbose("--> port num. unmatch: %d\n",
						ap->i_Port);
					match = 0;
				}
				Verbose("#### AUTHMATCH %s@%s --> %d\n",
					ap->i_user,ap->i_Host,match);
				if( match )
					break;
			}
		}

ACCUM:
		if( match && op == ON_EQ  || !match && op == ON_NEQ )
			do_onoff = DO_ON;
		else
		if( match && op == OFF_EQ || !match && op == OFF_NEQ )
			do_onoff = DO_OFF;
		else	do_onoff = DO_NOP;

		switch( do_onoff ){
			case DO_ON:	isin = 1; break;
			case DO_OFF:	isin = 0; break;
		}
		Verbose("[%d/%d] ==> %d (%s %s)\n",
			hi,nhosts,isin,hostlist->hl_what,hostname);
	}
	return isin ? hi : 0;
}
static VA_hostIsinList(hostlist,proto,cname,alias,Vaddr,username,lev)
	HostList *hostlist;
	char *proto,*cname,*alias;
	VAddr *Vaddr;
	char *username;
{
	return
	VA_hostIsinListX(hostlist,proto,cname,alias,Vaddr,username,0,NULL,lev);
}
hostIsinList(hostlist,proto,hostname,portnum,username)
	HostList *hostlist;
	char *proto,*hostname;
	char *username;
{
	return hostIsinListX(hostlist,proto,hostname,portnum,username,0,NULL);
}
hostIsinListX(hostlist,proto,hostname,portnum,username,ac,av)
	HostList *hostlist;
	char *proto,*hostname;
	char *username;
	AuthInfo *av[];
{	VAddr Vaddr;
	char *vdom,vdomb[256],hostnameb[512],*rhostname;
	char primaryname[512];
	char *vhost;
	VAddr Vaddrasis;

	if( vdom = scanVdomain(hostname,hostnameb,vdomb) )
		rhostname = hostnameb;
	else	rhostname = hostname;

	if( hostname[0] == '-' && !isMYSELF(hostname) ){
		vhost = hostname + 1;
		hostname += 1;
		VA_strtoVAddr(hostname,&Vaddr); /* AddrNull is set for non addr */
		/*
		Vaddr = AddrNull;
		*/
	}else{
		vhost = rhostname;
		primaryname[0] = 0;
		VA_gethostVAddr(0,rhostname,primaryname,&Vaddr);

		if( hostlist->hl_flags & HL_BYADDR )
		if( VA_strtoVAddr(rhostname,&Vaddrasis) ){
			Vaddr.a_ints = Vaddrasis.a_ints;
		}
		if( primaryname[0] ){
			hostname = primaryname;
			if( vdom ) strcat(hostname,vdom);
		}
	}
	Vaddr.a_port = portnum;
	return VA_hostIsinListX(hostlist,proto,hostname,vhost,&Vaddr,username,ac,av,0);
	/*
	return VA_hostIsinList(hostlist,proto,hostname,vhost,&Vaddr,username,0);
	*/
}
pathIsinList(hostlist,path)
	HostList *hostlist;
	char *path[];
{
}

static addUSER1(user,uv,uc)
	char *user,*uv[];
	int *uc;
{
	uv[*uc] = user;
	*uc += 1;
	return 0;
}
static addStrVec1(str,sv,sc)
	char *str,*sv[];
	int *sc;
{
	sv[*sc] = str;
	*sc += 1;
	return 0;
}

static scan_addrrange(Hp,addr)
	Host *Hp;
	char *addr;
{	char abuf[128],*ap,*dp;
	int ai,iaddr,lo,hi;

	strcpy(abuf,addr);
	ap = abuf;
	for( ai = 0; ai < 4;  ){
		if( dp = strchr(ap,'.') )
			*dp++ = 0;
		if( *ap == '[' ){
			if( sscanf(ap+1,"%d-%d",&lo,&hi) != 2 )
				goto error;
		}else	lo = hi = atoi(ap);

		Hp->ranges[ai].h_low = lo;
		Hp->ranges[ai].h_high = hi;
		ai++;
		if( dp == 0 )
			break;
		else	ap = dp;
	}
	for(; ai < 4; ai++ ){
		Hp->ranges[ai].h_low = 0;
		Hp->ranges[ai].h_high  = 255;
	}
	return 0;
error:
	sv1log("ERROR address range syntax: %s\n",addr);
	return -1;
}

static addHL1(hostmask,hostlist,op,users)
	char *hostmask;
	HostList *hostlist;
	char *users;
{	char op_hostmask[2048];

	if( users )
		sprintf(op_hostmask,"%c%s@%s",op,users,hostmask);
	else	sprintf(op_hostmask,"%c%s",op,hostmask);
	addHostList1(op_hostmask,hostlist);
	return 0;
}
static char *hostpart(hostmask,split)
	char *hostmask;
{	char *dp;

	if( dp = strchr(hostmask,'@') )
	if( dp == hostmask || dp[-1] != '/' ){ /* NOT host/@Cn */
		if( split )
			*dp = 0;
		return dp + 1;
	}
	return 0;
}

static addHL(hostmask,hostlist)
	char *hostmask;
	HostList *hostlist;
{	char list[2048],*dp;
	int op;
	char *hosts,*users;

	if( strchr(opsyms,*hostmask) )
		op = *hostmask++;
	else	op = '+';
	strcpy(list,hostmask);

	if( dp = hostpart(list,1) ){
		hosts = dp;
		users = list;
	}else{
		hosts = list;
		users = 0;
	}

/*
	if( streq(hosts,".localnet") ){
		strcpy(list,DELEGATE_LOCALNET);
		hosts = list;
	}else
*/
	if( *hosts != '{' )
		return 0;

	Verbose("addHL: %c [ %s ] @ [ %s ]\n",op,users?users:"*",hosts);
	scan_commaListL(hosts,0,addHL1,hostlist,op,users);
	return 1;
}

extern char **dupv();

addHostList1(hostmask,hostlist)
	char *hostmask;
	HostList *hostlist;
{	int idx,siz,nsiz;
	Host *Hp,**oHlist,**nHlist;

	if( strcmp(hostmask,hostmatch_asisaddr) == 0 ){
		hostlist->hl_flags |= HL_BYADDR;
		return 0;
	}
	if( strcmp(hostmask,hostmatch_withauth) == 0 ){
		hostlist->hl_flags |= HL_BYAUTH;
		return 0;
	}
	if( hostmask[0] == '!' && hostmask[1] == '!' )
		hostmask += 2;
	if( hostmask[0] == '+' && hostmask[1] == '!' )
		hostmask += 1;
	if( addHL(hostmask,hostlist) )
		return 0;
	if( hostlist->hl_cnt == 0 && *hostmask == '!' )
		if( strcmp(hostmask,"!*") != 0 )
		addHostList1("*",hostlist);

	Hp = NewStruct(Host);
	idx = ++hostlist->hl_cnt;
	siz = hostlist->hl_size;
	if( siz <= idx ){
		oHlist = hostlist->hl_List;
		if( hostlist->hl_inc )
			nsiz = siz + hostlist->hl_inc;
		else	nsiz = siz + HL_INCSIZE;
		while( nsiz <= idx )
			nsiz++;
		hostlist->hl_size = nsiz;
		if( oHlist == (Host**)0 )
			nHlist = (Host**)calloc(nsiz,sizeof(Host*));
		else	nHlist = (Host**)realloc(oHlist,nsiz*sizeof(Host*));
		hostlist->hl_List = nHlist;
	}
	hostlist->hl_List[idx] = Hp;
	makeHost(hostlist,Hp,0,hostmask);
	return 0;
}

canonWildcardAddr(hostname)
	char *hostname;
{	char *hp,ch,*bp,buff[128],*ap,addr[128];
	int noct,octv,olen,isdg;

	if( strchr(hostname,'*') && strchr(hostname,'.') ){
		hp = hostname;
		ap = addr;
		noct = 0;

		while( *hp ){
			bp = buff;
			isdg = 1;
			while( ch = *hp ){
				hp++;
				*bp++ = ch;
				if( ch == '.' ) 
					break;
				if( !isdigit(ch) )
					isdg = 0;
			}
			*bp = 0;

			if( buff[0]=='*' && (buff[1]==0||buff[1]=='.') ){
				sprintf(ap,"[0-255]%c",buff[1]);
			}else
			if( isdigit(buff[0]) && isdg ){
				octv = atoi(buff);
				if( octv < 0 || 255 < octv )
					goto NOTADDR;
				strcpy(ap,buff);
			}else	goto NOTADDR;
			ap += strlen(ap);

			noct++;
			if( 4 <= noct && *hp != 0 )
				goto NOTADDR;
		}
		if( 2 <= noct && noct <= 4 ){
			*ap = 0;
			strcpy(hostname,addr);
		}
	} NOTADDR:;
}

static isnetmasklen(mask)
	char *mask;
{	int masklen;

	if( isdigit(mask[0]) && mask[1] == 0
	 || isdigit(mask[0]) && isdigit(mask[1]) && mask[2] == 0 ){
		masklen = atoi(mask);
		if( 1 <= masklen && masklen <= 31 )
			return masklen;
	}
	return 0;
}

makeHost(hostlist,Hp,lev,hostmask)
	HostList *hostlist;
	Host *Hp;
	char *hostmask;
{	char *mp,hostnameb[1024],primaryname[512];
	VAddr addr,mask;
	int submask,mask1;
	char proto[256],hostmaskb[1024];
	char hostmaskb2[1024],vdomain[256];
	char *route;
	int listid;

	init_hostListTab();

	proto[0] = 0;
	if( strstr(hostmask,"://") ){
		sscanf(hostmask,"%[^:]://%s",proto,hostmaskb);
		hostmask = hostmaskb;
	}
	if( route = strstr(hostmask,"//") ){
		*route = 0;
		Hp->h_route = (Host*)calloc(sizeof(Host),1);
		makeHost(hostlist,Hp->h_route,lev+1,route+2);
	}

	if( hostmask[0] == '&' && hostmask[1] == 0 ){
		Hp->h_op = OP_AND;
		return;
	}
	if( hostmask[0] == '|' && hostmask[1] == 0 ){
		Hp->h_op = OP_OR;
		return;
	}
	if( hostmask[0] == '!' && hostmask[1] == 0 ){
		Hp->h_op = OP_NEG;
		return;
	}

	if( hostmask[0] == '!' && hostmask[1] == '!' )
		hostmask += 2;

	if( *hostmask == '!' ){ Hp->h_op = OFF_EQ;  hostmask++; }else
/*
	if( *hostmask == '!' ){ Hp->h_op = ON_NEQ;  hostmask++; }else
	if( *hostmask == '?' ){ Hp->h_op = OFF_NEQ; hostmask++; }else
	if( *hostmask == '-' ){ Hp->h_op = OFF_EQ;  hostmask++; }else
*/
			      { Hp->h_op = ON_EQ;
				if(*hostmask=='+')  hostmask++; }

	vdomain[0] = 0;
	if( scanVdomain(hostmask,hostmaskb2,vdomain) )
		hostmask = hostmaskb2;

	if( *hostmask == '-' ){
		int mtype = 0;
		char *dp;
		for( dp = hostmask+1; *dp; dp += 2 ){
			if( strneq(dp,"*/",2) ) mtype |= HT_BYANY; else
			if( strneq(dp,"A/",2) ) mtype |= HT_BYAGENT; else
			if( strneq(dp,"a/",2) ) mtype |= HT_BYAUTH; else
			if( strneq(dp,"c/",2) ) mtype |= HT_BYCERT; else
			if( strneq(dp,"h/",2) ) mtype |= HT_BYHOST; else break;
		}
		if( mtype ){
			hostmask = dp;
			Hp->h_type |= mtype;
		}
	}

	if( (hostlist->hl_flags & HL_BYAGENT) || (Hp->h_type & HT_BYAGENT) ){
		Hp->h_name = StrAlloc(hostmask);
		return;
	}

	mask = MaskZero;
	if( mp = strchr(hostmask,'/') ){
		*mp++ = 0;
		if( TimePeriod(hostmask) ){
			mask.I3 = atoi(mp);
		}else
		if( submask = isnetmasklen(mp) ){
			mask.I3 = 0xFFFFFFFF << (32-submask);
		}else
		if( strcmp(mp,".") == 0 ){
			mask.I3 = HEURISTIC_MASK;
		}else
		if( strchr(mp,'.') ) /* address in dot-notaion */
			VA_strtoVAddr(mp,&mask);
		else
		if( mp[0] == '@' ){
			switch( mp[1] ){
			    case 'A': mask1 = 0xFF000000; break;
			    case 'B': mask1 = 0xFFFF0000; break;
			    case 'C': mask1 = 0xFFFFFF00; break;
			    default:  mask1 = CLASS_MASK; break;
			}
			mask.I3 = mask1;
			if( submask = atoi(&mp[2]) )
				mask.I3 = mask.I3 >> submask;
		}else{
			sscanf(mp,"%x",&mask.I3);
		}
	}else{
		mask.I3 = 0xFFFFFFFF;
	}
	Hp->h_mask = mask;

	strcpy(hostnameb,hostmask);

	if( streq(hostmask,".") ){
		char me[128];
		GetHostname(me,sizeof(me));
		strcpy(hostnameb,me);
	}

	{	char *dp,*uv[128],users[2048];
		int ui;

		if( dp = strchr(hostnameb,'@') ){
			if( hostnameb[0] == '{' ){
				char *pp;
				strcpy(users,hostnameb+1);
				if( pp = strchr(users,'}') )
					*pp = 0;
				else	*strchr(users,'@') = 0;
			}else{
				strcpy(users,hostnameb);
				*strchr(users,'@') = 0;
			}
			ui = 0;
			scan_commaList(users,2,addUSER1,uv,&ui);
			uv[ui] = 0;
			Hp->h_user = dupv(uv,0);
			Hp->h_userRexp = makeUserRexp(ui,uv);

			strcpy(hostnameb,dp+1);

			if( (Hp->h_type&HT_BYANY)==0 || (Hp->h_type&HT_BYHOST) )
			if( !hostlist->hl_noIdent )
			if( hostnameb[0] != '!' )
				enableClientIdent(hostnameb);

			if( hostnameb[0] == '!' ){
				ovstrcpy(hostnameb,hostnameb+1);
				if( Hp->h_op == ON_EQ )
					Hp->h_op = OFF_EQ;
				else	Hp->h_op = ON_EQ;
			}
			/* USER part of USER@!host should be ignored ?? */
		}

		/* if no explicit "-I/" prefixed */
		if( (Hp->h_type & HT_BYANY) == 0 ){
			if( Hp->h_user == 0 )
				Hp->h_type |= HT_BYHOST;
			else	Hp->h_type |= HT_BYANY;
		}
	}
	if( HostIncomingIFX(hostnameb) ){
		if( hostnameb[2] && strchr(hostnameb+2,':') == 0 ){
			Strins(hostnameb+2,"*:"); /* -Pnum -> -P*:num */
		}
	}
	{	char *dp,*pv[128],ports[2048];
		int pi;

		if( dp = strchr(hostnameb,':') ){
			*dp++ = 0;
			strcpy(ports,dp);
			pi = 0;
			scan_commaList(ports,2,addStrVec1,pv,&pi);
			pv[pi] = 0;
			Hp->h_port = dupv(pv,0);
		}
	}
	if( proto[0] != 0 ){
		char *pv[128],protos[2048];
		int pi;

		pi = 0;
		strcpy(protos,proto);
		scan_commaList(protos,2,addStrVec1,pv,&pi);
		pv[pi] = 0;
		Hp->h_proto = dupv(pv,0);
	}

	primaryname[0] = 0;
	canonWildcardAddr(hostnameb);

	if( 0 < (listid = strid(hostListTabHash,hostnameb,-1)) ){
		Verbose("%s[%d]: %c HOSTLIST[%d]:%s(%s)\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op],listid,hostnameb,hostListName(listid));
		Hp->h_listid = listid;
	}else
	if( HostOnClientside(hostnameb) ){
		Verbose("%s[%d]: %c (ClientHost) = -C / %x\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op],Hp->h_mask.I3);
	}else
	if( HostIncomingIFX(hostnameb) ){
		ovstrcpy(hostnameb,hostnameb+2);
		Verbose("%s[%d]: %c (IncomingIFX) = -P%s / %x\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op],hostnameb,Hp->h_mask.I3);
		Hp->h_type |= HT_CLIF;
	}else
	if( HostIncomingIF(hostnameb) ){
		Verbose("%s[%d]: %c (IncomingIF) = %s / %x\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op],hostnameb,Hp->h_mask.I3);
	}else
	if( HostOutgoingIF(hostnameb) ){
		Verbose("%s[%d]: %c (OutgoingIF) = %s / %x\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op],hostnameb,Hp->h_mask.I3);
	}else
	if( TimePeriod(hostnameb) ){
		Verbose("%s[%d]: %c (TimePeriod) = %s / %d\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op],hostnameb,Hp->h_mask.I3);
	}else
	if( hostnameb[0] == '-' && !isMYSELF(hostnameb) ){
		Verbose("%s[%d]: %c %s\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op], hostnameb);
		if( VA_strtoVAddr(hostnameb+1,&addr) )
			Hp->h_addr = addr;
		else{
			Hp->h_asis = 1;

			if( hostlist->hl_what != LexHosts )
				addHostList1(hostnameb,lexHosts);

			ovstrcpy(hostnameb,hostnameb+1);
			/* to suppress resolution on reload which will
			 * always result in 'unknown' */
			if( !VA_gethostVAddr(0,hostnameb,primaryname,NULL) ){
				sv1log("Virtual Host Name: '%s'\n",hostnameb);
				sethostcache_predef(hostnameb);
			}
		}
	}else
	if( strchr(hostnameb,'[') ){
		if( scan_addrrange(Hp,hostnameb) == 0 )
		Verbose("%s[%d]: %c %s\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op],hostnameb);
	}else
	if( strchr(hostnameb,'*') || streq(hostnameb,"?") ){
		Verbose("%s[%d]: %c %s\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op], hostnameb);
	}else
	if( !VA_gethostVAddr(0,hostnameb,primaryname,&addr) ){
		sv1log("ERROR %s[%d] %s ? unknown\n",
			hostlist->hl_what,hostlist->hl_cnt, hostnameb);
	}else{
		if( primaryname[0] )
			strcpy(hostnameb,primaryname);
		Hp->h_addr = addr;
		default_netmask(&Hp->h_mask,&addr);

		Verbose("%s[%d]: %c %s = %x / %x\n",
			hostlist->hl_what,hostlist->hl_cnt,
			opsyms[Hp->h_op], hostnameb,
			addr.I3,Hp->h_mask.I3);
	}

	Hp->h_name = StrAlloc(hostnameb);
	Hp->h_vdomain = vdomain[0] ? StrAlloc(vdomain) : 0;
}


/*
 * A name/address to be compared its identity only with its lexical
 * representation without resolved (possibly unified) name/address.
 * When a host is referred as such one (prefixed with "-" like "-host"
 * or "-10.0.0.1"), it should be compared lexically anytime.
 */
hostcmp_lexical(h1,h2,cacheonly)
	char *h1,*h2;
{
	/*
	if( VA_hostIsinList(lexHosts,"*",h1,h1,&AddrNull,"",0) )
	if( VA_hostIsinList(lexHosts,"*",h2,h2,&AddrNull,"",0) ){
	*/
	if( VA_hostIsinList(lexHosts,"*",h1,h1,&AddrNull,"",0)
	 || VA_hostIsinList(lexHosts,"*",h2,h2,&AddrNull,"",0)
	){
		return strcasecmp(h1,h2);
	}
	if( cacheonly )
		return hostcmp_incache(h1,h2);
	else	return hostcmp(h1,h2);
}

typedef struct {
	char	 *name;
	char	 *path;
	HostList *hlist;
} PathList; 
static PathList *pathlists;
static int pathlists_siz;
static int pathlists_cnt;

int makePathList(what,path)
	char *what,*path;
{	HostList *HL;
	PathList *PL;
	int nsize,hi;

	if( pathlists_siz <= pathlists_cnt ){
		pathlists_siz += 32;
		nsize = sizeof(PathList) * pathlists_siz;
		pathlists = (PathList*)Malloc(pathlists,nsize);
	}
	for( hi = 0; hi < pathlists_cnt; hi++ )
		if( strcmp(path,pathlists[hi].path) == 0 )
			return hi+1;

	HL = NewStruct(HostList);
	HL->hl_what = StrAlloc(what);

	scan_commaListL(path,0,addHostList1,HL);

	PL = &pathlists[pathlists_cnt++];
	PL->hlist = HL;
	PL->path = StrAlloc(path);

	return pathlists_cnt;
}
matchPathList(hlid,path)
	char *path;
{
}
matchPath1(hlid,user,host,port)
	char *user,*host;
{	HostList *HL;
	int match;

	if( hlid < 1 ){
		sv1log("##FATAL: matchPath1(%d) %s@%s:%d\n",hlid,user,host,port);
		return 0;
	}
	HL = pathlists[hlid-1].hlist;
	match = hostIsinList(HL,"*",host,port,user);
Verbose("###### matchPath1(%d,%s,%s:%d) = %d\n",hlid,user,host,port,match);
	return match;
}

char *topofPath1(hlid,hostport,rbuff)
	char *hostport,*rbuff;
{	HostList *HL;
	Host *Hp;
	char *dp,*host,hostb[128];
	VAddr va;

	va = AddrNull;
	if( strchr(hostport,':') ){
		host = hostb;
		dp = wordscanY(hostport,hostb,sizeof(hostb),"^:");
		if( *dp == ':' )
			va.a_port = atoi(dp+1);
	}else	host = hostport;

	HL = pathlists[hlid-1].hlist;
	if( VA_hostIsinList(HL,"*",host,host,&va,"",0) )
		return hostport;

	Hp = HL->hl_List[1];
	if( Hp->h_port == 0 )
		return Hp->h_name;
	sprintf(rbuff,"%s:%d",Hp->h_name,atoi(Hp->h_port[0]));
	return rbuff;
}
